import numpy as np
import cykdtree as kdtree


def tree(method, pts, left_edge, right_edge, periodic, *args, **kwargs):
    r"""Get tree for a given domain decomposition schema.

    Args:
        method (str): Domain decomposition method. Supported options are:
            'kdtree': KDTree based on median position along the dimension
                with the greatest domain width. See
                :meth:`cgal4py.domain_decomp.kdtree` for details on
                accepted keyword arguments.
        pts (np.ndarray of float64): (n, m) array of n coordinates in a
            m-dimensional domain.
        left_edge (np.ndarray of float64): (m,) domain minimum in each
            dimension.
        right_edge (np.ndarray of float64): (m,) domain maximum in each
            dimension.
        *args: Variable argument list. Passed to the selected domain
            decomposition method.
        **kwargs: Variable keyword argument list. Passed to the selected
            domain decomposition method.

    Returns:
        object: Tree of type specified by `method`.

    Raises:
        ValueError: If `method` is not one of the supported domain
            decomposition methods listed above.

    """
    # Get leaves
    if method.lower() == 'kdtree':
        tree = kdtree.PyKDTree(pts, left_edge, right_edge, *args, **kwargs)
    else:
        raise ValueError("'{}' is not a supported ".format(method) +
                         "domain decomposition.")
    # Return tree
    return tree


class GenericLeaf(object):
    def __init__(self, npts, left_edge, right_edge):
        r"""A generic container for leaf info with the minimum required info.
        These leaves must still be processed to add additional properties using
        :meth:`cgal4py.domain_decomp.process_leaves`, but can serve as a base
        class for supplemental domain decompositions and be provided to
        :class:`cgal4py.domain_decomp.GenericTree`.

        Args:
            npts (int): Number of points on this leaf.
            left_edge (np.ndarray of float64): Leaf min along each dimension.
            right_edge (np.ndarray of float64): Leaf max along each dimension.

        Attributes:
            npts (int): Number of points on this leaf, including those added
                during communication.
            left_edge (np.ndarray of float64): Domain minimum along each
                dimension.
            right_edge (np.ndarray of float64): Domain maximum along each
                dimension.

        """
        self.npts = npts
        self.left_edge = left_edge
        self.right_edge = right_edge

    @classmethod
    def from_leaf(cls, leaf):
        r"""Construct a GenericLeaf from a non-generic leaf.

        Args:
            leaf (Leaf): A leaf object.
       
        Returns:
            :class:`cgal4py.domain_decomp.GenericLeaf`: Generic version of the
                input leaf.

        """
        out = cls(leaf.npts, leaf.left_edge, leaf.right_edge)
        other_attr = ['id', 'ndim', 'num_leaves', 'start_idx', 'stop_idx',
                      'domain_width', 'periodic_left', 'periodic_right',
                      'left_neighbors', 'right_neighbors', 'neighbors']
        for k in other_attr:
            if hasattr(leaf, k):
                setattr(out, k, getattr(leaf, k))
        return out


class GenericTree(object):
    def __init__(self, idx, leaves, left_edge, right_edge, periodic):
        r"""Generic container for domain decomposition with the minimal
        required info. The leaves must have at least the following
        attributes:
            npts (int): Number of points on the leaf.
            left_edge (np.ndarray of float): min of leaf extent in each
                dimension.
            right_edge (np.ndarray of float): max of leaf extent in each
                dimension.

        Args:
            idx (np.ndarray of int): Indices sorting points in the tree by the
                leaf that contains them.
            leaves (list of leaf objects): Leaves in an arbitrary domain
                decomposition.
            left_edge (np.ndarray of float): domain minimum in each dimension.
            right_edge (np.ndarray of float): domain maximum in each dimension.
            periodic (bool): True if domain is periodic, False otherwise.

        Attributes:
            idx (np.ndarray of int): Indices sorting points in the tree by the
                leaf that contains them.
            leaves (list of leaf objects): Leaves in an arbitrary domain
                decomposition.
            left_edge (np.ndarray of float): domain minimum in each dimension.
            right_edge (np.ndarray of float): domain maximum in each dimension.
            periodic (bool): True if domain is periodic, False otherwise.
            num_leaves (int): Number of leaves in the tree.
            domain_width (np.ndarray of float): Domain width in each dimension.

        """
        self.idx = idx
        self.left_edge = left_edge
        self.right_edge = right_edge
        self.periodic = periodic
        self.domain_width = right_edge - left_edge
        self.num_leaves = len(leaves)
        self.leaves = process_leaves(leaves, left_edge, right_edge, periodic)

    @classmethod
    def from_tree(cls, tree):
        r"""Construct a GenericTree from a non-generic tree.

        Args:
            tree (Tree): A tree object.
       
        Returns:
            :class:`cgal4py.domain_decomp.GenericTree`: Generic version of the
                input tree.

        """
        leaves = [GenericLeaf.from_leaf(leaf) for leaf in tree.leaves]
        out = cls(tree.idx, leaves, tree.left_edge, tree.right_edge,
                  tree.periodic)
        other_attr = []
        for k in other_attr:
            if hasattr(leaf, k):
                setattr(out, k, getattr(leaf, k))
        return out


def process_leaves(leaves, left_edge, right_edge, periodic):
    r"""Process leaves ensuring they have the necessary information/methods
    for parallel tessellation. Each leaf must have at least the following
    attributes:
        npts (int): Number of points on the leaf.
        left_edge (np.ndarray of float): min of leaf extent in each dimension.
        right_edge (np.ndarray of float): max of leaf extent in each dimension.

    Args:
        leaves (list of leaf objects): Leaves in an arbitrary domain
            decomposition.
        left_edge (np.ndarray of float64): domain minimum in each dimension.
        right_edge (np.ndarray of float64): domain maximum in each dimension.
        periodic (bool): True if domain is periodic, False otherwise.

    Returns:
        list of leaf objects: Leaves process with additional attributes added
            if they do not exist and can be added.

    Raises:
        AttributeError: If a leaf does not have one of the required attributes.
        TypeError: If a leaf has a required attribute, but of the wrong type.
        ValueError: If a leaf has a required attribute, but of the wrong size.

    """
    ndim = len(left_edge)
    req_attr = {'npts': [int, long, np.int32, np.uint32, np.int64, np.uint64],
                'left_edge': ([float, np.float32, np.float64], (ndim,)),
                'right_edge': ([float, np.float32, np.float64], (ndim,))}
    # Check for necessary attributes
    for k, v in req_attr.items():
        if isinstance(v, tuple):
            for i, leaf in enumerate(leaves):
                if not hasattr(leaf, k):
                    raise AttributeError(
                        "Leaf {} does not have attribute {}.".format(i, k))
                lv = getattr(leaf, k)
                if not isinstance(lv, np.ndarray):
                    raise TypeError("Attribute {} ".format(k) +
                                    "of leaf {} ".format(i) +
                                    "is not an array.\n" +
                                    "It is type {}.".format(type(lv)))
                if lv.dtype not in v[0]:
                    raise TypeError("Attribute {} ".format(k) +
                                    "of leaf {} ".format(i) +
                                    "is not an array with dtype " +
                                    "{}.\n".format(v[0]) +
                                    "It is type {}.".format(lv.dtype))
                if v[1] is not None and lv.shape != v[1]:
                    raise ValueError("Attribute {} ".format(k) +
                                     "of leaf {} ".format(i) +
                                     "is not an array with shape " +
                                     "{}.\n".format(v[1]) +
                                     "It is shape {}.".format(lv.shape))
        else:
            for i, leaf in enumerate(leaves):
                if not hasattr(leaf, k):
                    raise AttributeError("Leaf {} does not ".format(i) +
                                         "have attribute {}.".format(k))
                lv = getattr(leaf, k)
                if not isinstance(lv, tuple(v)):
                    raise TypeError("Attribute {} ".format(k) +
                                    "of leaf {} is not ".format(i) +
                                    "of type {}.\n".format(v) +
                                    "It is type {}.".format(type(lv)))
    # Set id & ensure leaves are sorted
    if getattr(leaves[0], 'id', None) is None:
        for i, leaf in enumerate(leaves):
            leaf.id = i
    else:
        leaves = sorted(leaves, key=lambda l: l.id)
    # Set number of dimensions
    if getattr(leaves[0], 'ndim', None) is None:
        for leaf in leaves:
            leaf.ndim = ndim
    # Set total number of leaves
    if getattr(leaves[0], 'num_leaves', None) is None:
        num_leaves = len(leaves)
        for leaf in leaves:
            leaf.num_leaves = num_leaves
    # Set starting index
    if getattr(leaves[0], 'start_idx', None) is None:
        nprev = 0
        for leaf in leaves:
            leaf.start_idx = nprev
            nprev += leaf.npts
    # Set stopping index
    if getattr(leaves[0], 'stop_idx', None) is None:
        for leaf in leaves:
            leaf.stop_idx = leaf.start_idx + leaf.npts
    # Set domain width
    if getattr(leaves[0], 'domain_width', None) is None:
        domain_width = right_edge - left_edge
        for leaf in leaves:
            leaf.domain_width = domain_width
    # Determine if leaves are on periodic boundaries
    if getattr(leaves[0], 'periodic_left', None) is None:
        if periodic:
            for leaf in leaves:
                leaf.periodic_left = np.isclose(leaf.left_edge, left_edge)
                leaf.periodic_right = np.isclose(leaf.right_edge, right_edge)
        else:
            for leaf in leaves:
                leaf.periodic_left = np.zeros(leaf.ndim, 'bool')
                leaf.periodic_right = np.zeros(leaf.ndim, 'bool')
    # Add neighbors
    if getattr(leaves[0], 'left_neighbors', None) is None:
        for j, leaf in enumerate(leaves):
            leaf.left_neighbors = [[] for _ in range(ndim)]
            leaf.right_neighbors = [[] for _ in range(ndim)]
            for prev in leaves[:(j+1)]:
                match = True
                for i in range(ndim):
                    if leaf.left_edge[i] > prev.right_edge[i]:
                        if not (leaf.periodic_right[i] and
                                prev.periodic_left[i]):
                            match = False
                            break
                    if leaf.right_edge[i] < prev.left_edge[i]:
                        if not (prev.periodic_right[i] and
                                leaf.periodic_left[i]):
                            match = False
                            break
                if match:
                    for i in range(ndim):
                        if np.isclose(leaf.left_edge[i], prev.right_edge[i]):
                            leaf.left_neighbors[i].append(prev.id)
                            prev.right_neighbors[i].append(leaf.id)
                        elif np.isclose(leaf.right_edge[i], prev.left_edge[i]):
                            leaf.right_neighbors[i].append(prev.id)
                            prev.left_neighbors[i].append(leaf.id)
                        if periodic:
                            if (leaf.periodic_right[i] and
                                    prev.periodic_left[i]):
                                leaf.right_neighbors[i].append(prev.id)
                                prev.left_neighbors[i].append(leaf.id)
                            if (prev.periodic_right[i] and
                                    leaf.periodic_left[i]):
                                leaf.left_neighbors[i].append(prev.id)
                                prev.right_neighbors[i].append(leaf.id)
    if getattr(leaves[0], 'neighbors', None) is None:
        for leaf in leaves:
            neighbors = [leaf.id]
            for i in range(ndim):
                neighbors += leaf.left_neighbors[i]
                neighbors += leaf.right_neighbors[i]
            leaf.neighbors = list(set(neighbors))
    # Return leaves
    return leaves


__all__ = ["tree", "kdtree", "GenericLeaf", "GenericTree", "process_leaves"]



"""
delaunayD.pyx

Wrapper for CGAL 3D Delaunay Triangulation
"""

import cython

import numpy as np
cimport numpy as np
import struct

from cgal4py import plot
from cgal4py.delaunay.tools cimport sortSerializedTess

from libc.stdlib cimport malloc, free
from libcpp.vector cimport vector
from libcpp.set cimport set as cset
from libcpp.pair cimport pair
from libcpp cimport bool as cbool
from cpython cimport bool as pybool
from cython.operator cimport dereference
from cython.operator cimport preincrement, predecrement
from libc.stdint cimport uint32_t, uint64_t, int32_t, int64_t

ctypedef uint32_t info_t
cdef object np_info = np.uint32
ctypedef np.uint32_t np_info_t

def is_valid():
    if (VALID == 1):
        return True
    else:
        return False

cdef class DelaunayD_vertex:
    r"""Wrapper class for a triangulation vertex.

    Attributes:
        T (:obj:`Delaunay_with_info_D[info_t]`): C++ Triangulation object 
            that this vertex belongs to. 
        x (:obj:`Delaunay_with_info_D[info_t].Vertex`): C++ vertex object 
            Direct interaction with this object is not recommended. 

    """
    cdef Delaunay_with_info_D[info_t] *T
    cdef Delaunay_with_info_D[info_t].Vertex x

    cdef void assign(self, Delaunay_with_info_D[info_t] *T,
                     Delaunay_with_info_D[info_t].Vertex x):
        r"""Assign C++ objects to attributes.

            Args:
            T (:obj:`Delaunay_with_info_D[info_t]`): C++ Triangulation object 
                that this vertex belongs to. 
            x (:obj:`Delaunay_with_info_D[info_t].Vertex`): C++ vertex object 
                Direct interaction with this object is not recommended. 

        """
        self.T = T
        self.x = x

    @property
    def num_dims(self):
        r"""int: Number of dimensions."""
        return D

    def __repr__(self):
        cdef str out = "DelaunayD_vertex[{} at ".format(self.index)
        cdef np.ndarray[np.float64_t] p = self.point
        for i in range(self.num_dims-1):
            out += "{:+7.2e},".format(p[i])
        out += "{:+7.2e}]".format(p[self.num_dims-1])
        return out

    def __richcmp__(DelaunayD_vertex self, DelaunayD_vertex solf, int op):
        if (op == 2):
            return <pybool>(self.x == solf.x)
        elif (op == 3):
            return <pybool>(self.x != solf.x)
        else:
            raise NotImplementedError

    def is_infinite(self):
        r"""Determine if the vertex is the infinite vertex.
        
        Returns:
            bool: True if the vertex is the infinite vertex, False otherwise.

        """
        return self.T.is_infinite(self.x)

    def set_point(self, np.ndarray[np.float64_t, ndim=1] pos):
        r"""Set this vertex's corrdinates.

        Args:
            pos (:obj:`ndarray` of float64): new coordinates for vertex.

        """
        self.T.updated = <cbool>True
        assert(len(pos) == self.num_dims)
        self.x.set_point(&pos[0])

    def set_cell(self, DelaunayD_cell c):
        r"""Set the designated cell for this vertex.

        Args:
            c (DelaunayD_cell): Cell that should be set as the designated cell.

        """
        self.T.updated = <cbool>True
        self.x.set_cell(c.x)

    property point:
        r""":obj:`ndarray` of :obj:`float64`: The cartesian coordinates of the
        vertex."""
        def __get__(self):
            cdef np.ndarray[np.float64_t] out
            out = np.zeros(D, 'float64')
            if self.is_infinite():
                out[:] = np.float('inf')
            else:
                self.x.point(&out[0])
            return out

    property index:
        r"""info_t: The index of the vertex point in the input array."""
        def __get__(self):
            cdef info_t out
            if self.is_infinite():
                out = <info_t>np.iinfo(np_info).max
            else:
                out = self.x.info()
            return out

    property dual_volume:
        r"""float64: The volume of the dual Voronoi cell. If the volume is 
        infinite, -1.0 is returned."""
        def __get__(self):
            cdef np.float64_t out = self.T.dual_volume(self.x)
            return out

    property cell:
        r"""DelaunayD_cell: Designated cell for this vertex."""
        def __get__(self):
            cdef Delaunay_with_info_D[info_t].Cell c
            c = self.x.cell()
            cdef DelaunayD_cell out = DelaunayD_cell()
            out.assign(self.T, c)
            return out

    def incident_vertices(self):
        r"""Find vertices that are adjacent to this vertex.

        Returns:
            DelaunayD_vertex_vector: Iterator over vertices incident to this 
                vertex.

        """
        cdef vector[Delaunay_with_info_D[info_t].Vertex] it
        it = self.T.incident_vertices(self.x)
        cdef DelaunayD_vertex_vector out = DelaunayD_vertex_vector()
        out.assign(self.T, it)
        return out

    def incident_faces(self, int i):
        r"""Find faces that are incident to this vertex.

        Args:
            int i: The dimensionality of the faces that should be returned.

        Returns:
            DelaunayD_face_vector: Iterator over faces incident to this
                vertex.

        """
        cdef vector[Delaunay_with_info_D[info_t].Face] it
        it = self.T.incident_faces(self.x, i)
        cdef DelaunayD_face_vector out = DelaunayD_face_vector()
        out.assign(self.T, it)
        return out

    def incident_cells(self):
        r"""Find cells that are incident to this vertex.

        Returns:
            DelaunayD_cell_vector: Iterator over cells incident to this vertex.

        """
        cdef vector[Delaunay_with_info_D[info_t].Cell] it
        it = self.T.incident_cells(self.x)
        cdef DelaunayD_cell_vector out = DelaunayD_cell_vector()
        out.assign(self.T, it)
        return out


cdef class DelaunayD_vertex_iter:
    r"""Wrapper class for a triangulation vertex iterator.

    Args:
        T (DelaunayD): Triangulation that this vertex belongs to.
        vert (:obj:`str`, optional): String specifying the vertex that 
            should be referenced. Valid options include: 
                'all_begin': The first vertex in an iteration over all vertices.  
                'all_end': The last vertex in an iteration over all vertices. 
 
    Attributes:
        T (:obj:`Delaunay_with_info_D[info_t]`): C++ Triangulation object 
            that this vertex belongs to. 
        x (:obj:`Delaunay_with_info_D[info_t].All_verts_iter`): C++ vertex 
            object. Direct interaction with this object is not recommended. 

    """
    cdef Delaunay_with_info_D[info_t] *T
    cdef Delaunay_with_info_D[info_t].All_verts_iter x

    def __cinit__(self, DelaunayD T, str vert = None):
        self.T = T.T
        if vert == 'all_begin':
            self.x = self.T.all_verts_begin()
        elif vert == 'all_end':
            self.x = self.T.all_verts_end()

    def __richcmp__(DelaunayD_vertex_iter self, DelaunayD_vertex_iter solf, 
                    int op):
        if (op == 2):
            return <pybool>(self.x == solf.x)
        elif (op == 3):
            return <pybool>(self.x != solf.x)
        else:
            raise NotImplementedError

    def is_infinite(self):
        r"""Determine if the vertex is the infinite vertex.
        
        Returns:
            bool: True if the vertex is the infinite vertex, False otherwise.

        """
        return self.T.is_infinite(self.x)

    def increment(self):
        r"""Advance to the next vertex in the triangulation."""
        preincrement(self.x)

    def decrement(self):
        r"""Advance to the previous vertex in the triangulation."""
        predecrement(self.x)

    property vertex:
        r"""DelaunayD_vertex: Corresponding vertex object."""
        def __get__(self):
            cdef DelaunayD_vertex out = DelaunayD_vertex()
            out.assign(self.T, Delaunay_with_info_D[info_t].Vertex(self.x)) 
            return out


cdef class DelaunayD_vertex_range:
    r"""Wrapper class for iterating over a range of triangulation vertices

    Args:
        xstart (DelaunayD_vertex_iter): The starting vertex.  
        xstop (DelaunayD_vertex_iter): Final vertex that will end the iteration. 
        finite (:obj:`bool`, optional): If True, only finite verts are 
            iterated over. Otherwise, all verts are iterated over. Defaults 
            False.

    Attributes:
        x (DelaunayD_vertex_iter): The current vertex. 
        xstop (DelaunayD_vertex_iter): Final vertex that will end the iteration. 
        finite (bool): If True, only finite verts are iterater over. Otherwise
            all verts are iterated over. 

    """
    cdef DelaunayD_vertex_iter x
    cdef DelaunayD_vertex_iter xstop
    cdef pybool finite
    def __cinit__(self, DelaunayD_vertex_iter xstart, 
                  DelaunayD_vertex_iter xstop,
                  pybool finite = False):
        self.x = xstart
        self.xstop = xstop
        self.finite = finite

    def __iter__(self):
        return self

    def __next__(self):
        cdef DelaunayD_vertex out
        if self.finite:
            while (self.x != self.xstop) and self.x.is_infinite():
                self.x.increment()
        if self.x != self.xstop:
            out = self.x.vertex
            self.x.increment()
            return out
        else:
            raise StopIteration()

cdef class DelaunayD_vertex_vector:
    r"""Wrapper class for a vector of vertices.

    Attributes:
        T (:obj:`Delaunay_with_info_D[info_t]`): C++ triangulation object.
            Direct interaction with this object is not recommended.
        v (:obj:`vector[Delaunay_with_info_D[info_t].Vertex]`): Vector of C++ 
            vertices.
        n (int): The number of vertices in the vector.
        i (int): The index of the currect vertex.

    """
    cdef Delaunay_with_info_D[info_t] *T
    cdef vector[Delaunay_with_info_D[info_t].Vertex] v
    cdef int n
    cdef int i

    cdef void assign(self, Delaunay_with_info_D[info_t] *T,
                     vector[Delaunay_with_info_D[info_t].Vertex] v):
        r"""Assign C++ attributes.

        Args:
            T (:obj:`Delaunay_with_info_D[info_t]`): C++ triangulation object.
                Direct interaction with this object is not recommended.
            v (:obj:`vector[Delaunay_with_info_D[info_t].Vertex]`): Vector of 
                C++ vertices.

        """
        self.T = T
        self.v = v
        self.n = v.size()
        self.i = 0

    def __iter__(self):
        return self

    def __next__(self):
        cdef DelaunayD_vertex out
        if self.i < self.n:
            out = DelaunayD_vertex()
            out.assign(self.T, self.v[self.i])
            self.i += 1
            return out
        else:
            raise StopIteration()

    def __getitem__(self, i):
        cdef DelaunayD_vertex out
        if isinstance(i, int):
            out = DelaunayD_vertex()
            out.assign(self.T, self.v[i])
            return out
        else:
            raise TypeError("DelaunayD_vertex_vector indices must be itegers, "+
                            "not {}".format(type(i)))


cdef class DelaunayD_face:
    r"""Wrapper class for a triangulation face.

    Attributes:
        T (:obj:`Delaunay_with_info_D[info_t]`): C++ Triangulation object 
            that this face belongs to. 
        x (:obj:`Delaunay_with_info_D[info_t].Face`): C++ face object 
            Direct interaction with this object is not recommended. 

    """
    cdef Delaunay_with_info_D[info_t] *T
    cdef Delaunay_with_info_D[info_t].Face x

    cdef void assign(self, Delaunay_with_info_D[info_t] *T,
                     Delaunay_with_info_D[info_t].Face x):
        r"""Assign C++ objects to attributes.

            Args:
            T (:obj:`Delaunay_with_info_D[info_t]`): C++ Triangulation object 
                that this face belongs to. 
            x (:obj:`Delaunay_with_info_D[info_t].Face`): C++ face object 
                Direct interaction with this object is not recommended. 

        """
        self.T = T
        self.x = x

    @property
    def dim(self):
        r"""int: Number of dimensions this face occupies."""
        return self.x.dim()

    @property
    def nverts(self):
        r"""int: Number of vertices in this face (self.ndim+1)."""
        return self.dim+1

    def __repr__(self):
        cdef str out = "DelaunayD_face[{}d: ".format(self.dim)
        cdef int i
        for i in range(self.nverts-1):
            out += "{},".format(repr(self.vertex(i)))
        out += "{}]".format(repr(self.vertex(self.nverts-1)))
        return out

    # def __richcmp__(DelaunayD_face self, DelaunayD_face solf, int op):
    #     if (op == 2):
    #         return <pybool>(self.x == solf.x)
    #     elif (op == 3):
    #         return <pybool>(self.x != solf.x)
    #     else:
    #         raise NotImplementedError

    def is_infinite(self):
        r"""Determine if the face is incident to the infinite vertex.
        
        Returns:
            bool: True if the face is incident to the infinite vertex, False 
                otherwise.

        """
        return self.T.is_infinite(self.x)

    def is_equivalent(self, DelaunayD_face solf):
        r"""Determine if another face has the same vertices as this face.

        Args:
            solf (DelaunayD_face): Face for comparison.

        Returns:
            bool: True if the two faces share the same vertices, False
                otherwise.

        """
        return <pybool>self.T.are_equal(self.x, solf.x)

    def vertex(self, int i):
        r"""Get the ith vertex on this face.

        Args:
            i (int): Index of vertex to return.

        Returns:
            DelaunayD_vertex: ith vertex on this face.

        """
        cdef Delaunay_with_info_D[info_t].Vertex x
        x = self.x.vertex(i)
        cdef DelaunayD_vertex out = DelaunayD_vertex()
        out.assign(self.T, x)
        return out

    def index(self, int i):
        r"""Get the index in the cell of the ith vertex on this face.

        Args:
            i (int): Index of vertex on this face to get index for.
        
        Returns:
            int: Index of ith vertex on this face within the reference cell.

        """
        cdef int out
        out = self.x.ind(i)
        return out

    property cell:
        r"""DelaunayD_cell: The cell this face is assigned to."""
        def __get__(self):
            cdef Delaunay_with_info_D[info_t].Cell c
            c = self.x.cell()
            cdef DelaunayD_cell out = DelaunayD_cell()
            out.assign(self.T, c)
            return out

    property center:
        r""":obj:`ndarray` of float64: coordinates of face center."""
        def __get__(self):
            if self.is_infinite():
                return np.float('inf')*np.ones(3, 'float64')
            else:
                ptot = self.vertex(0).point
                for i in range(1,self.nverts):
                    ptot += self.vertex(i).point
                return ptot/self.nverts

    property area:
        r"""float64: The area of the face. If infinite, -1 is returned"""
        def __get__(self):
            # return self.T.n_simplex_volume(self.x)
            raise NotImplementedError

    def incident_vertices(self):
        r"""Find vertices that are incident to this face.

        Returns:
            DelaunayD_vertex_vector: Iterator over vertices incident to this 
                face.

        """
        cdef vector[Delaunay_with_info_D[info_t].Vertex] it
        it = self.T.incident_vertices(self.x)
        cdef DelaunayD_vertex_vector out = DelaunayD_vertex_vector()
        out.assign(self.T, it)
        return out

    def incident_faces(self, int i):
        r"""Find faces that are incident to this face.

        Args:
            int i: Dimensionality of faces that should be returned.

        Returns:
            DelaunayD_face_vector: Iterator over faces incident to this face. 

        """
        cdef vector[Delaunay_with_info_D[info_t].Face] it
        it = self.T.incident_faces(self.x, i)
        cdef DelaunayD_face_vector out = DelaunayD_face_vector()
        out.assign(self.T, it)
        return out

    def incident_cells(self):
        r"""Find cells that are incident to this face.

        Returns:
            DelaunayD_cell_vector: Iterator over cells incident to this face.

        """
        cdef vector[Delaunay_with_info_D[info_t].Cell] it
        it = self.T.incident_cells(self.x)
        cdef DelaunayD_cell_vector out = DelaunayD_cell_vector()
        out.assign(self.T, it)
        return out


cdef class DelaunayD_face_vector:
    r"""Wrapper class for a vector of faces.

    Attributes:
        T (:obj:`Delaunay_with_info_D[info_t]`): C++ triangulation object.
            Direct interaction with this object is not recommended.
        v (:obj:`vector[Delaunay_with_info_D[info_t].Face]`): Vector of C++ 
            faces.
        n (int): The number of faces in the vector.
        i (int): The index of the currect face.

    """
    cdef Delaunay_with_info_D[info_t] *T
    cdef vector[Delaunay_with_info_D[info_t].Face] v
    cdef int n
    cdef int i

    cdef void assign(self, Delaunay_with_info_D[info_t] *T,
                     vector[Delaunay_with_info_D[info_t].Face] v):
        r"""Assign C++ attributes.

        Args:
            T (:obj:`Delaunay_with_info_D[info_t]`): C++ triangulation object.
                Direct interaction with this object is not recommended.
            v (:obj:`vector[Delaunay_with_info_D[info_t].Face]`): Vector of 
                C++ faces.

        """
        self.T = T
        self.v = v
        self.n = v.size()
        self.i = 0

    def __iter__(self):
        return self

    def __next__(self):
        cdef DelaunayD_face out
        if self.i < self.n:
            out = DelaunayD_face()
            out.assign(self.T, self.v[self.i])
            self.i += 1
            return out
        else:
            raise StopIteration()

    def __getitem__(self, i):
        cdef DelaunayD_face out
        if isinstance(i, int):
            out = DelaunayD_face()
            out.assign(self.T, self.v[i])
            return out
        else:
            raise TypeError("DelaunayD_face_vector indices must be itegers, "+
                            "not {}".format(type(i)))


cdef class DelaunayD_facet:
    r"""Wrapper class for a triangulation facet.

    Attributes:
        T (:obj:`Delaunay_with_info_D[info_t]`): C++ Triangulation object 
            that this facet belongs to. 
        x (:obj:`Delaunay_with_info_D[info_t].Facet`): C++ facet object 
            Direct interaction with this object is not recommended. 

    """
    cdef Delaunay_with_info_D[info_t] *T
    cdef Delaunay_with_info_D[info_t].Facet x

    cdef void assign(self, Delaunay_with_info_D[info_t] *T,
                     Delaunay_with_info_D[info_t].Facet x):
        r"""Assign C++ objects to attributes.

            Args:
            T (:obj:`Delaunay_with_info_D[info_t]`): C++ Triangulation object 
                that this facet belongs to. 
            x (:obj:`Delaunay_with_info_D[info_t].Facet`): C++ facet object 
                Direct interaction with this object is not recommended. 

        """
        self.T = T
        self.x = x

    @property
    def dim(self):
        r"""int: Number of dimensions that the facet covers (one less than
        overall domain)."""
        return D - 1

    @property
    def nverts(self):
        r"""int: Number of vertices in the facet (same as number of dimensions
        in the domain."""
        return D

    @staticmethod
    def from_cell(DelaunayD_cell c, int i):
        r"""Construct a facet from a cell and index of the vertex in the cell 
        opposite the desired facet.

        Args:
            c (DelaunayD_cell): Cell 
            i (int): Index of vertex in c that is opposite the facet.

        Returns:
            DelaunayD_facet: Facet incident to c and opposite vertex i in c.

        """
        cdef DelaunayD_facet out = DelaunayD_facet()
        cdef Delaunay_with_info_D[info_t].Facet e
        e = Delaunay_with_info_D[info_t].Facet(c.x, i)
        out.assign(c.T, e)
        return out

    def __repr__(self):
        cdef str out = "DelaunayD_facet["
        cdef int i
        for i in range(self.nverts-1):
            out += "{},".format(repr(self.vertex(i)))
        out += "{}]".format(repr(self.vertex(self.nverts-1)))
        return out

    # def __richcmp__(DelaunayD_facet self, DelaunayD_facet solf, int op):
    #     if (op == 2):
    #         return <pybool>(self.x == solf.x)
    #     elif (op == 3):
    #         return <pybool>(self.x != solf.x)
    #     else:
    #         raise NotImplementedError

    def is_infinite(self):
        r"""Determine if the facet is incident to the infinite vertex.
        
        Returns:
            bool: True if the facet is incident to the infinite vertex, False 
                otherwise.

        """
        return self.T.is_infinite(self.x)

    def is_equivalent(self, DelaunayD_facet solf):
        r"""Determine if another facet has the same vertices as this facet.

        Args:
            solf (DelaunayD_facet): Facet for comparison.

        Returns:
            bool: True if the two facets share the same vertices, False 
                otherwise.

        """
        return <pybool>self.T.are_equal(self.x, solf.x)

    def vertex(self, int i):
        r"""Get the ith vertex incident to this facet.

        Args:
            i (int): Index of vertex that should be returned.

        Returns:
            Delaunay_vertex: ith vertex of this facet.

        """
        cdef Delaunay_with_info_D[info_t].Vertex v
        v = self.x.vertex(i)
        cdef DelaunayD_vertex out = DelaunayD_vertex()
        out.assign(self.T, v)
        return out

    property center:
        r""":obj:`ndarray` of float64: x,y,z coordinates of cell center."""
        def __get__(self):
            if self.is_infinite():
                return np.float('inf')*np.ones(self.dim, 'float64')
            else:
                ptot = self.vertex(0).point
                for i in range(1,self.nverts):
                    ptot += self.vertex(i).point
                return ptot/self.nverts
                        
    property area:
        r"""float64: The area of the facet. If infinite, -1 is returned"""
        def __get__(self):
            # return self.T.n_simplex_volume(self.x)
            raise NotImplementedError

    property cell:
        r"""DelaunayD_cell: The cell this facet is assigned to."""
        def __get__(self):
            cdef Delaunay_with_info_D[info_t].Cell c
            c = self.x.cell()
            cdef DelaunayD_cell out = DelaunayD_cell()
            out.assign(self.T, c)
            return out

    property ind:
        r"""int: The index of the vertex this facet is opposite on its cell."""
        def __get__(self):
            return self.x.ind()


cdef class DelaunayD_facet_vector:
    r"""Wrapper class for a vector of facets.

    Attributes:
        T (:obj:`Delaunay_with_info_D[info_t]`): C++ triangulation object.
            Direct interaction with this object is not recommended.
        v (:obj:`vector[Delaunay_with_info_D[info_t].Facet]`): Vector of C++ 
            facets.
        n (int): The number of facets in the vector.
        i (int): The index of the currect facet.

    """
    cdef Delaunay_with_info_D[info_t] *T
    cdef vector[Delaunay_with_info_D[info_t].Facet] v
    cdef int n
    cdef int i

    cdef void assign(self, Delaunay_with_info_D[info_t] *T,
                     vector[Delaunay_with_info_D[info_t].Facet] v):
        r"""Assign C++ attributes.

        Args:
            T (:obj:`Delaunay_with_info_D[info_t]`): C++ triangulation object.
                Direct interaction with this object is not recommended.
            v (:obj:`vector[Delaunay_with_info_D[info_t].Facet]`): Vector of 
                C++ facets.

        """
        self.T = T
        self.v = v
        self.n = v.size()
        self.i = 0

    def __iter__(self):
        return self

    def __next__(self):
        cdef DelaunayD_facet out
        if self.i < self.n:
            out = DelaunayD_facet()
            out.assign(self.T, self.v[self.i])
            self.i += 1
            return out
        else:
            raise StopIteration()

    def __getitem__(self, i):
        cdef DelaunayD_facet out
        if isinstance(i, int):
            out = DelaunayD_facet()
            out.assign(self.T, self.v[i])
            return out
        else:
            raise TypeError("DelaunayD_facet_vector indices must be itegers, "+
                            "not {}".format(type(i)))


cdef class DelaunayD_cell:
    r"""Wrapper class for a triangulation cell.

    Attributes:
        T (:obj:`Delaunay_with_info_D[info_t]`): C++ Triangulation object 
            that this cell belongs to. 
        x (:obj:`Delaunay_with_info_D[info_t].Cell`): C++ cell object. 
            Direct interaction with this object is not recommended.

    """
    cdef Delaunay_with_info_D[info_t] *T
    cdef Delaunay_with_info_D[info_t].Cell x

    @property
    def dim(self):
        r"""int: Number of dimensions that the facet covers (one less than
        overall domain)."""
        return self.T.num_dims()

    @property
    def nverts(self):
        r"""int: Number of vertices in the facet (same as number of dimensions
        in the domain."""
        return self.dim + 1

    cdef void assign(self, Delaunay_with_info_D[info_t] *T,
                     Delaunay_with_info_D[info_t].Cell x):
        r"""Assign C++ objects to attributes.

            Args:
            T (:obj:`Delaunay_with_info_D[info_t]`): C++ Triangulation object 
                that this cell belongs to. 
            x (:obj:`Delaunay_with_info_D[info_t].Cell`): C++ cell object 
                Direct interaction with this object is not recommended. 

        """
        self.T = T
        self.x = x

    def __repr__(self):
        cdef str out = "Delaunay2_cell["
        cdef int i
        for i in range(self.nverts-1):
            out += "{},".format(repr(self.vertex(i)))
        out += "{}]".format(repr(self.vertex(self.nverts-1)))
        return out

    def __richcmp__(DelaunayD_cell self, DelaunayD_cell solf, int op):
        if (op == 2):
            return <pybool>(self.x == solf.x)
        elif (op == 3):
            return <pybool>(self.x != solf.x)
        else:
            raise NotImplementedError

    def is_infinite(self):
        r"""Determine if the cell is incident to the infinite vertex.

        Returns:
            bool: True if the cell is incident to the infinite vertex, False 
                otherwise.

        """
        return self.T.is_infinite(self.x)

    def mirror_index(self, int i):
        r"""Get the index of this cell with respect to its ith neighbor. 

        Args: 
            i (int): Index of neighbor that should be used to determine the 
                mirrored index. 

        Returns: 
            int: Index of this cell with respect to its ith neighbor. 

        """
        cdef int out = self.T.mirror_index(self.x, i)
        return out

    def mirror_vertex(self, int i):
        r"""Get the vertex of this cell's ith neighbor that is opposite to this 
        cell.

        Args: 
            i (int): Index of neighbor that should be used to determine the 
                mirrored vertex. 

        Returns:
            DelaunayD_vertex: Vertex in the ith neighboring cell of this cell, 
                that is opposite to this cell. 

        """
        cdef Delaunay_with_info_D[info_t].Vertex vc
        vc = self.T.mirror_vertex(self.x, i)
        cdef DelaunayD_vertex out = DelaunayD_vertex()
        out.assign(self.T, vc)
        return out

    def facet(self, int i):
        r"""Find the facet opposite the ith vertex incident to this cell.

        Args:
            i (int): Index of vertex opposite the desired facet.

        Returns:
            DelaunayD_facet: The facet opposite the ith vertex incident to this 
                cell.

        """
        cdef DelaunayD_facet out = DelaunayD_facet.from_cell(self, i)
        return out

    def vertex(self, int i):
        r"""Find the ith vertex that is incident to this cell. 

        Args:
            i (int): The index of the vertex that should be returned.

        Returns:
            DelaunayD_vertex: The ith vertex incident to this cell. 

        """
        cdef Delaunay_with_info_D[info_t].Vertex v
        v = self.x.vertex(i)
        cdef DelaunayD_vertex out = DelaunayD_vertex()
        out.assign(self.T, v)
        return out

    def has_vertex(self, DelaunayD_vertex v, pybool return_index = False):
        r"""Determine if a vertex belongs to this cell.

        Args:
            v (DelaunayD_vertex): Vertex to test ownership for. 
            return_index (:obj:`bool`, optional): If True, the index of the 
                vertex within the cell is returned in the event that it is a 
                vertex of the cell. Otherwise, the index is not returned.  
        
        Returns:
            bool: True if the vertex is part of the cell, False otherwise. In
                the event that `return_index = True` and the vertex is a part of  
                the cell, an integer specifying the index of the vertex within
                the cell will be returned instead. 
        """
        cdef int i = -1
        cdef cbool out
        if return_index:
            out = self.x.has_vertex(v.x, &i)
            return i
        else:
            out = self.x.has_vertex(v.x)
            return <pybool>out
            
    def ind_vertex(self, DelaunayD_vertex v):
        r"""Determine the index of a vertex within a cell. 

        Args: 
            v (DelaunayD_vertex): Vertex to find index for. 

        Returns: 
            int: Index of vertex within the cell. 

        """
        return self.x.ind(v.x)

    def neighbor(self, int i):
        r"""Find the neighboring cell opposite the ith vertex of this cell. 

        Args: 
            i (int): The index of the neighboring cell that should be returned. 

        Returns: 
            Delaunay2_cell: The neighboring cell opposite the ith vertex. 

        """
        cdef Delaunay_with_info_D[info_t].Cell v
        v = self.x.neighbor(i)
        cdef DelaunayD_cell out = DelaunayD_cell()
        out.assign(self.T, v)
        return out

    def has_neighbor(self, DelaunayD_cell v, pybool return_index = False):
        r"""Determine if a cell is a neighbor to this cell. 

        Args: 
            v (DelaunayD_cell): Cell to test as a neighbor. 
            return_index (:obj:`bool`, optional): If True, the index of the 
                neighbor within the cell is returned in the event that it is a 
                neighbor of the cell. Otherwise, the index is not returned. 

        Returns: 
            bool: True if the given cell is a neighbor, False otherwise. In 
                the event that `return_index = True` and the v is a neighbor of 
                this cell, an integer specifying the index of the neighbor to 
                will be returned instead. 

        """
        cdef int i = -1
        cdef cbool out = self.x.has_neighbor(v.x, &i)
        if out and return_index:
            return i
        else:
            return <pybool>out

    def ind_neighbor(self, DelaunayD_cell v):
        r"""Determine the index of a neighboring cell. 

        Args: 
            v (DelaunayD_cell): Neighboring cell to find index for. 

        Returns: 
            int: Index of vertex opposite to neighboring cell. 

        """
        return self.x.ind(v.x)

    def set_vertex(self, int i, DelaunayD_vertex v):
        r"""Set the ith vertex of this cell. 

        Args: 
            i (int): Index of this cell's vertex that should be set. 
            v (DelaunayD_vertex): Vertex to set ith vertex of this cell to. 

        """
        self.T.updated = <cbool>True
        self.x.set_vertex(i, v.x)

    def set_neighbor(self, int i, DelaunayD_cell n):
        r"""Set the ith neighboring cell of this cell. 

        Args: 
            i (int): Index of this cell's neighbor that should be set. 
            n (DelaunayD_cell): Cell to set ith neighbor of this cell to. 

        """
        self.T.updated = <cbool>True
        self.x.set_neighbor(i, n.x)

    property center:
        """:obj:`ndarray` of float64: coordinates of cell center."""
        def __get__(self):
            cdef np.ndarray[np.float64_t] ptot
            cdef int i
            if self.is_infinite():
                return np.float('inf')*np.ones(self.dim, 'float64')
            else:
                ptot = self.vertex(0).point
                for i in range(1, self.nverts):
                    ptot += self.vertex(i).point
                return ptot/self.nverts

    property circumcenter:
        """:obj:`ndarray` of float64: coordinates of cell circumcenter."""
        def __get__(self):
            cdef np.ndarray[np.float64_t] out = np.zeros(self.dim, 'float64')
            self.T.circumcenter(self.x, &out[0])
            return out

    def incident_vertices(self):
        r"""Find vertices that are incident to this cell.

        Returns:
            DelaunayD_vertex_vector: Iterator over vertices incident to this 
                cell.

        """
        cdef vector[Delaunay_with_info_D[info_t].Vertex] it
        it = self.T.incident_vertices(self.x)
        cdef DelaunayD_vertex_vector out = DelaunayD_vertex_vector()
        out.assign(self.T, it)
        return out

    def incident_faces(self, int i):
        r"""Find faces that are incident to this cell.

        Args:
            i (int): Dimensionality of the faces that should be returned.

        Returns:
            DelaunayD_face_vector: Iterator over faces incident to this cell. 

        """
        cdef vector[Delaunay_with_info_D[info_t].Face] it
        it = self.T.incident_faces(self.x, i)
        cdef DelaunayD_face_vector out = DelaunayD_face_vector()
        out.assign(self.T, it)
        return out

    def incident_cells(self):
        r"""Find cells that are incident to this cell.

        Returns:
            DelaunayD_cell_vector: Iterator over cells incident to this cell.

        """
        cdef vector[Delaunay_with_info_D[info_t].Cell] it
        it = self.T.incident_cells(self.x)
        cdef DelaunayD_cell_vector out = DelaunayD_cell_vector()
        out.assign(self.T, it)
        return out


cdef class DelaunayD_cell_iter:
    r"""Wrapper class for a triangulation cell iteration.

    Args:
        T (DelaunayD): Triangulation that this cell belongs to. 
        cell (:obj:`str`, optional): String specifying the cell that
            should be referenced. Valid options include: 
                'all_begin': The first cell in an iteration over all cells. 
                'all_end': The last cell in an iteration over all cells.
    
    Attributes:
        T (:obj:`Delaunay_with_info_D[info_t]`): C++ Triangulation object 
            that this cell belongs to. 
        x (:obj:`Delaunay_with_info_D[info_t].All_cells_iter`): C++ cell
            object. Direct interaction with this object is not recommended.

    """
    cdef Delaunay_with_info_D[info_t] *T
    cdef Delaunay_with_info_D[info_t].All_cells_iter x

    def __cinit__(self, DelaunayD T, str cell = None):
        self.T = T.T
        if cell == 'all_begin':
            self.x = self.T.all_cells_begin()
        elif cell == 'all_end':
            self.x = self.T.all_cells_end()

    def __richcmp__(DelaunayD_cell_iter self, DelaunayD_cell_iter solf, int op):
        if (op == 2):
            return <pybool>(self.x == solf.x)
        elif (op == 3):
            return <pybool>(self.x != solf.x)
        else:
            raise NotImplementedError

    def is_infinite(self):
        r"""Determine if the cell is incident to the infinite vertex.

        Returns:
            bool: True if the cell is incident to the infinite vertex, False 
                otherwise.

        """
        return self.T.is_infinite(self.x)

    def increment(self):
        r"""Advance to the next cell in the triangulation."""
        preincrement(self.x)

    def decrement(self):
        r"""Advance to the previous cell in the triangulation."""
        predecrement(self.x)

    property cell:
        r"""DelaunayD_cell: Corresponding cell object."""
        def __get__(self):
            cdef DelaunayD_cell out = DelaunayD_cell()
            out.T = self.T
            out.x = Delaunay_with_info_D[info_t].Cell(self.x)
            return out


cdef class DelaunayD_cell_range:
    r"""Wrapper class for iterating over a range of triangulation cells.

    Args:
        xstart (DelaunayD_cell_iter): The starting cell. 
        xstop (DelaunayD_cell_iter): Final cell that will end the iteration. 
        finite (:obj:`bool`, optional): If True, only finite cells are 
            iterated over. Otherwise, all cells are iterated over. Defaults
            to False.  

    Attributes:
        x (DelaunayD_cell_iter): The current cell. 
        xstop (DelaunayD_cell_iter): Final cell that will end the iteration. 
        finite (bool): If True, only finite cells are iterated over. Otherwise, 
            all cells are iterated over.   

    """
    cdef DelaunayD_cell_iter x
    cdef DelaunayD_cell_iter xstop
    cdef pybool finite
    def __cinit__(self, DelaunayD_cell_iter xstart, DelaunayD_cell_iter xstop,
                  pybool finite = False):
        self.x = xstart
        self.xstop = xstop
        self.finite = finite

    def __iter__(self):
        return self

    def __next__(self):
        if self.finite:
            while (self.x != self.xstop) and self.x.is_infinite():
                self.x.increment()
        cdef DelaunayD_cell out
        if self.x != self.xstop:
            out = self.x.cell
            self.x.increment()
            return out
        else:
            raise StopIteration()

cdef class DelaunayD_cell_vector:
    r"""Wrapper class for a vector of cells.

    Attributes:
        T (:obj:`Delaunay_with_info_D[info_t]`): C++ triangulation object.
            Direct interaction with this object is not recommended.
        v (:obj:`vector[Delaunay_with_info_D[info_t].Cell]`): Vector of C++ 
            cells.
        n (int): The number of cells in the vector.
        i (int): The index of the currect cell.

    """
    cdef Delaunay_with_info_D[info_t] *T
    cdef vector[Delaunay_with_info_D[info_t].Cell] v
    cdef int n
    cdef int i

    cdef void assign(self, Delaunay_with_info_D[info_t] *T,
                     vector[Delaunay_with_info_D[info_t].Cell] v):
        r"""Assign C++ attributes.

        Args:
            T (:obj:`Delaunay_with_info_D[info_t]`): C++ triangulation object.
                Direct interaction with this object is not recommended.
            v (:obj:`vector[Delaunay_with_info_D[info_t].Cell]`): Vector of 
                C++ cells.

        """
        self.T = T
        self.v = v
        self.n = v.size()
        self.i = 0

    def __iter__(self):
        return self

    def __next__(self):
        cdef DelaunayD_cell out
        if self.i < self.n:
            out = DelaunayD_cell()
            out.T = self.T
            out.x = self.v[self.i]
            self.i += 1
            return out
        else:
            raise StopIteration()

    def __getitem__(self, i):
        cdef DelaunayD_cell out
        if isinstance(i, int):
            out = DelaunayD_cell()
            out.assign(self.T, self.v[i])
            return out
        else:
            raise TypeError("DelaunayD_cell_vector indices must be itegers, "+
                            "not {}".format(type(i)))


cdef class DelaunayD:
    r"""Wrapper class for a 3D Delaunay triangulation.

    Attributes:
        n (int): The number of points inserted into the triangulation.
        T (:obj:`Delaunay_with_info_D[info_t]`): C++ triangulation object. 
            Direct interaction with this object is not recommended. 
        n_per_insert (list of int): The number of points inserted at each
            insert.

    """

    cdef Delaunay_with_info_D[info_t] *T
    cdef readonly int n
    cdef public object n_per_insert
    cdef readonly pybool _locked
    cdef public object _cache_to_clear_on_update

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __cinit__(self):
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            self.T = new Delaunay_with_info_D[info_t]()
        self.n = 0
        self.n_per_insert = []
        self._locked = False
        self._cache_to_clear_on_update = {}

    def _lock(self):
        self._locked = True
    def _unlock(self):
        self._locked = False
    def _set_updated(self):
        self.T.updated = <cbool>True

    def _update_tess(self):
        if self.T.updated:
            self._cache_to_clear_on_update.clear()
            self.T.updated = <cbool>False

    @staticmethod
    def _update_to_tess(func):
        def wrapped_func(solf, *args, **kwargs):
            solf._lock()
            out = func(solf, *args, **kwargs)
            solf._unlock()
            solf._set_updated()
            solf._update_tess()
            return out
        return wrapped_func

    @staticmethod
    def _dependent_property(fget):
        attr = '_'+fget.__name__
        def wrapped_fget(solf):
            if solf._locked:
                raise RuntimeError("Cannot get dependent property '{}' while triangulation is locked.".format(attr))
            solf._update_tess()
            if attr not in solf._cache_to_clear_on_update:
                solf._cache_to_clear_on_update[attr] = fget(solf)
            return solf._cache_to_clear_on_update[attr]
        return property(wrapped_fget, None, None, fget.__doc__)

    def is_valid(self):
        r"""Determine if the triangulation is a valid Delaunay triangulation. 

        Returns: 
            bool: True if the triangulation is valid, False otherwise. 

        """
        cdef cbool out
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            out = self.T.is_valid()
        return <pybool>out

    def is_equivalent(DelaunayD self, DelaunayD solf):
        r"""Determine if two triangulations are equivalent. Currently only 
        checks that the triangulations have the same numbers of vertices, cells,
        edges, and facets.

        Args:
            solf (:class:`cgal4py.delaunay.DelaunayD`): Triangulation this one 
                should be compared to.

        Returns:
            bool: True if the two triangulations are equivalent.

        """
        cdef cbool out
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            out = self.T.is_equal(dereference(solf.T))
        return <pybool>(out)

    @classmethod
    def from_file(cls, fname):
        r"""Create a triangulation from one saved to a file.

        Args:
            fname (str): Full path to file where triangulation is saved.

        Returns:
            :class:`cgal4py.delaunay.DelaunayD`: Triangulation constructed from
                saved information.

        """
        T = cls()
        T.read_from_file(fname)
        return T

    @classmethod
    def from_serial(cls, *args):
        r"""Create a triangulation from serialized information. 

        Args: 
            *args: All arguments are passed to :meth:`cgal4py.delaunay.DelaunayD.deserialize`. 

        Returns: 
            :class:`cgal4py.delaunay.DelaunayD`: Triangulation constructed from 
                deserialized information. 

        """
        T = cls()
        T.deserialize(*args)
        return T

    @classmethod
    def from_serial_buffer(cls, *args, **kwargs):
        r"""Create a triangulation from serialized information in a buffer.

        Args:
            See :meth:`cgal4py.delaunay.Delaunay3.deserialize_from_buffer`.

        Returns:
            :class:`cgal4py.delaunay.Delaunay3`: Triangulation constructed from
                deserialized information.

        """
        T = cls()
        T.deserialize_from_buffer(*args, **kwargs)
        return T

    def serialize_to_buffer(self, buf, pos=None):
        r"""Write serialized triangulation to a buffer.

        Args:
            buf (file): File buffer.
            pos (np.ndarray, optional): Positions to be written. If not
                provided, positions are not included and must be provided
                during any subsequent read. Defaults to None.

        """
        cdef np.ndarray[np_info_t, ndim=2] cells
        cdef np.ndarray[np_info_t, ndim=2] neighbors
        cdef info_t idx_inf
        cells, neighbors, idx_inf = self.serialize()
        cdef str ifmt, ffmt
        ifmt = cells.dtype.char
        ffmt = 'd'
        if pos is not None:
            ffmt = pos.dtype.char
        buf.write(struct.pack('cc', ffmt, ifmt))
        buf.write(struct.pack(ifmt, idx_inf))
        if pos is not None:
            buf.write(struct.pack(2*ifmt, pos.shape[0], pos.shape[1]))
            buf.write(pos.tobytes())
        buf.write(struct.pack(2*ifmt, cells.shape[0], cells.shape[1]))
        buf.write(cells.tobytes())
        buf.write(neighbors.tobytes())

    def deserialize_from_buffer(self, buf, pos=None):
        r"""Read a serialized triangulation from the buffer.

        Args:
            buf (file): File buffer.
            pos (np.ndarray, optional): Positions to be used for deserializing
                the triangulation if the positions are not in the file. If
                not provided, the file is assumed to contain the positions.
                Defaults to None.

       """
        cdef int ndim = D
        cdef np.ndarray[np_info_t, ndim=2] cells
        cdef np.ndarray[np_info_t, ndim=2] neighbors
        cdef info_t idx_inf
        cdef str ifmt, ffmt
        cdef int isiz, fsiz
        cdef int nx, ny
        (ffmt, ifmt) = struct.unpack('cc', buf.read(struct.calcsize('cc')))
        fsiz = struct.calcsize(ffmt)
        isiz = struct.calcsize(ifmt)
        (idx_inf,) = struct.unpack(ifmt, buf.read(isiz))
        nx, ny = struct.unpack(2*ifmt, buf.read(2*isiz))
        assert(ny == ndim)
        pos = np.frombuffer(
            bytearray(buf.read(nx*ny*fsiz)), dtype=np.dtype(ffmt),
            count=nx*ny).reshape(nx, ny)
        nx, ny = struct.unpack(2*ifmt, buf.read(2*isiz))
        assert(ny == (ndim+1))
        cells = np.frombuffer(
            bytearray(buf.read(nx*ny*isiz)), dtype=np.dtype(ifmt),
            count=nx*ny).reshape(nx, ny)
        neigh = np.frombuffer(
            bytearray(buf.read(nx*ny*isiz)), dtype=np.dtype(ifmt),
            count=nx*ny).reshape(nx, ny)
        self.deserialize(pos, cells, neigh, idx_inf)

    # def write_to_file(self, fname):
    #     r"""Write the serialized tessellation information to a file. 

    #     Args:
    #         fname (str): The full path to the file that the tessellation should 
    #             be written to.

    #     """
    #     cdef char* cfname = fname
    #     with nogil, cython.boundscheck(False), cython.wraparound(False):
    #         self.T.write_to_file(cfname)

    # @_update_to_tess
    # def read_from_file(self, fname):
    #     r"""Read serialized tessellation information from a file.

    #     Args:
    #         fname (str): The full path to the file that the tessellation should 
    #             be read from.

    #     """
    #     cdef char* cfname = fname
    #     with nogil, cython.boundscheck(False), cython.wraparound(False):
    #         self.T.read_from_file(cfname)
    #     self.n = self.T.num_finite_verts()
    #     self.n_per_insert.append(self.n)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def serialize(self, pybool sort = False):
        r"""Serialize triangulation. 

        Args:
            sort (bool, optional): If True, cells info is sorted so that the 
                verts are in descending order for each cell and ascending order 
                overall. Defaults to False.

        Returns: 
            tuple containing:
                cells (np.ndarray of info_t): (n,m) Indices for m vertices in 
                    each of the n cells. A value of np.iinfo(np_info).max 
                    indicates the infinite vertex. 
                neighbors (np.ndarray of info_t): (n,l) Indices in `cells` of 
                    the m neighbors to each of the n cells. 
                idx_inf (I): Value representing the infinite vertex and or 
                    a missing neighbor.

        """
        cdef info_t n, m, i
        cdef int32_t d, j
        cdef np.ndarray[np_info_t, ndim=2] cells
        cdef np.ndarray[np_info_t, ndim=2] neighbors
        # Initialize arrays based on properties
        n = self.T.num_finite_verts()
        m = self.T.num_cells()
        assert(n == self.num_finite_verts)
        assert(m == self.num_cells)
        d = self.num_dims
        cells = np.zeros((m, d+1), np_info)
        neighbors = np.zeros((m, d+1), np_info)
        # Serialize and convert to original vertex order
        cdef info_t idx_inf
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            idx_inf = self.T.serialize_idxinfo[info_t](
                n, m, d, &cells[0,0], &neighbors[0,0])
        # Sort if desired
        if sort:
            with nogil, cython.boundscheck(False), cython.wraparound(False):
                sortSerializedTess[info_t](&cells[0,0], &neighbors[0,0],
                                           m, d+1)
        return cells, neighbors, idx_inf

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _serialize_info2idx_int32(self, info_t max_info,
                                  np.ndarray[np.int32_t] idx,
                                  pybool sort=False):
        cdef int32_t n, m
        cdef int32_t d
        cdef np.ndarray[np.int32_t, ndim=2] cells
        cdef np.ndarray[np.int32_t, ndim=2] neighbors
        n = self.T.num_finite_verts()
        m = self.T.num_cells()
        assert(idx.size >= n)
        d = self.num_dims
        cells = np.zeros((m, d+1), 'int32')
        neighbors = np.zeros((m, d+1), 'int32')
        cdef int32_t idx_inf
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            idx_inf = self.T.serialize_info2idx[int32_t](
                n, m, d, &cells[0,0], &neighbors[0,0], max_info, &idx[0])
        cells.resize(m, d+1, refcheck=False)
        neighbors.resize(m, d+1, refcheck=False)
        if sort:
            with nogil, cython.boundscheck(False), cython.wraparound(False):
                sortSerializedTess[int32_t](&cells[0,0], &neighbors[0,0],
                                            m, d+1)
        return cells, neighbors, idx_inf

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _serialize_info2idx_uint32(self, info_t max_info,
                                   np.ndarray[np.uint32_t] idx,
                                   pybool sort=False):
        cdef uint32_t n, m
        cdef int32_t d
        cdef np.ndarray[np.uint32_t, ndim=2] cells
        cdef np.ndarray[np.uint32_t, ndim=2] neighbors
        n = self.T.num_finite_verts()
        m = self.T.num_cells()
        assert(idx.size >= n)
        d = self.num_dims
        cells = np.zeros((m, d+1), 'uint32')
        neighbors = np.zeros((m, d+1), 'uint32')
        cdef uint32_t idx_inf
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            idx_inf = self.T.serialize_info2idx[uint32_t](
                n, m, d, &cells[0,0], &neighbors[0,0], max_info, &idx[0])
        cells.resize(m, d+1, refcheck=False)
        neighbors.resize(m, d+1, refcheck=False)
        if sort:
            with nogil, cython.boundscheck(False), cython.wraparound(False):
                sortSerializedTess[uint32_t](&cells[0,0], &neighbors[0,0],
                                             m, d+1)
        return cells, neighbors, idx_inf

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _serialize_info2idx_int64(self, info_t max_info,
                                  np.ndarray[np.int64_t] idx,
                                  pybool sort=False):
        cdef int64_t n, m
        cdef int32_t d
        cdef np.ndarray[np.int64_t, ndim=2] cells
        cdef np.ndarray[np.int64_t, ndim=2] neighbors
        n = self.T.num_finite_verts()
        m = self.T.num_cells()
        assert(idx.size >= n)
        d = self.num_dims
        cells = np.zeros((m, d+1), 'int64')
        neighbors = np.zeros((m, d+1), 'int64')
        cdef int64_t idx_inf
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            idx_inf = self.T.serialize_info2idx[int64_t](
                n, m, d, &cells[0,0], &neighbors[0,0], max_info, &idx[0])
        cells.resize(m, d+1, refcheck=False)
        neighbors.resize(m, d+1, refcheck=False)
        if sort:
            with nogil, cython.boundscheck(False), cython.wraparound(False):
                sortSerializedTess[int64_t](&cells[0,0], &neighbors[0,0],
                                            m, d+1)
        return cells, neighbors, idx_inf

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _serialize_info2idx_uint64(self, info_t max_info,
                                   np.ndarray[np.uint64_t] idx,
                                   pybool sort=False):
        cdef uint64_t n, m
        cdef int32_t d
        cdef np.ndarray[np.uint64_t, ndim=2] cells
        cdef np.ndarray[np.uint64_t, ndim=2] neighbors
        n = self.T.num_finite_verts()
        m = self.T.num_cells()
        assert(idx.size >= n)
        d = self.num_dims
        cells = np.zeros((m, d+1), 'uint64')
        neighbors = np.zeros((m, d+1), 'uint64')
        cdef uint64_t idx_inf
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            idx_inf = self.T.serialize_info2idx[uint64_t](
                n, m, d, &cells[0,0], &neighbors[0,0], max_info, &idx[0])
        cells.resize(m, d+1, refcheck=False)
        neighbors.resize(m, d+1, refcheck=False)
        if sort:
            with nogil, cython.boundscheck(False), cython.wraparound(False):
                sortSerializedTess[uint64_t](&cells[0,0], &neighbors[0,0],
                                             m, d+1)
        return cells, neighbors, idx_inf

    def serialize_info2idx(self, max_info, idx, pybool sort = False):
        r"""Serialize triangulation, only including some vertices and 
        translating the indices. 

        Args:
            max_info (info_t): Maximum value of info for verts that will be
                included in the serialization.  
            idx (np.ndarray of I): Indices that should be used to map from 
                vertex info.     
            sort (bool, optional): If True, cells info is sorted so that the 
                verts are in descending order for each cell and ascending order 
                overall. Defaults to False.

        Returns: 
            tuple containing:
                cells (np.ndarray of info_t): (n,m) Indices for m vertices in 
                    each of the n cells. A value of np.iinfo(np_info).max 
                    indicates the infinite vertex. 
                neighbors (np.ndarray of info_t): (n,l) Indices in `cells` of 
                    the m neighbors to each of the n cells. 
                idx_inf (I): Value representing the infinite vertex and or 
                    a missing neighbor.

        """
        if idx.dtype == np.int32:
            return self._serialize_info2idx_int32(<info_t>max_info, idx, sort)
        elif idx.dtype == np.uint32:
            return self._serialize_info2idx_uint32(<info_t>max_info, idx, sort)
        elif idx.dtype == np.int64:
            return self._serialize_info2idx_int64(<info_t>max_info, idx, sort)
        elif idx.dtype == np.uint64:
            return self._serialize_info2idx_uint64(<info_t>max_info, idx, sort)
        else:
            raise TypeError("idx.dtype = {} is not supported.".format(
                idx.dtype))

    @_update_to_tess
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def deserialize(self, np.ndarray[np.float64_t, ndim=2] pos,
                    np.ndarray[np_info_t, ndim=2] cells,
                    np.ndarray[np_info_t, ndim=2] neighbors,
                    info_t idx_inf):
        r"""Deserialize triangulation. 

        Args: 
            pos (np.ndarray of float64): Coordinates of points. 
            cells (np.ndarray of info_t): (n,m) Indices for m vertices in each 
                of the n cells. A value of np.iinfo(np_info).max A value of 
                np.iinfo(np_info).max indicates the infinite vertex. 
            neighbors (np.ndarray of info_t): (n,l) Indices in `cells` of the m 
                neighbors to each of the n cells. 
            idx_inf (info_t): Index indicating a vertex is infinite. 

        """
        cdef info_t n = pos.shape[0]
        cdef info_t m = cells.shape[0]
        cdef int32_t d = neighbors.shape[1]-1
        if (n == 0) or (m == 0):
            return
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            self.T.deserialize_idxinfo[info_t](n, m, d, &pos[0,0], 
                                               &cells[0,0], &neighbors[0,0],
                                               idx_inf)
        self.n = n
        self.n_per_insert.append(n)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @_update_to_tess
    def deserialize_with_info(self, np.ndarray[np.float64_t, ndim=2] pos,
                              np.ndarray[np_info_t, ndim=1] info,
                              np.ndarray[np_info_t, ndim=2] cells,
                              np.ndarray[np_info_t, ndim=2] neighbors,
                              info_t idx_inf):
        r"""Deserialize triangulation. 

        Args: 
            pos (np.ndarray of float64): Coordinates of points. 
            info (np.ndarray of info_t): Info for points. 
            cells (np.ndarray of info_t): (n,m) Indices for m vertices in each 
                of the n cells. A value of np.iinfo(np_info).max A value of 
                np.iinfo(np_info).max indicates the infinite vertex. 
            neighbors (np.ndarray of info_t): (n,l) Indices in `cells` of the m 
                neighbors to each of the n cells. 
            idx_inf (info_t): Index indicating a vertex is infinite. 

        """
        cdef info_t n = pos.shape[0]
        cdef info_t m = cells.shape[0]
        cdef int32_t d = neighbors.shape[1]-1
        if (n == 0) or (m == 0):
            return
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            self.T.deserialize[info_t](n, m, d, &pos[0,0], &info[0],
                                       &cells[0,0], &neighbors[0,0], idx_inf)
        self.n = n
        self.n_per_insert.append(n)

    @_dependent_property
    def num_dims(self):
        r"""int: The number of dimensions in the triangulation."""
        return self.T.num_dims()
    @_dependent_property
    def num_finite_verts(self): 
        r"""int: The number of finite vertices in the triangulation."""
        return self.T.num_finite_verts()
    @_dependent_property
    def num_finite_edges(self):
        r"""int: The number of finite edges in the triangulation."""
        return self.T.num_finite_faces(1)
    @_dependent_property
    def num_finite_facets(self):
        r"""int: The number of finite facets in the triangulation."""
        return self.T.num_finite_faces(D-1)
    @_dependent_property
    def num_finite_cells(self):
        r"""int: The number of finite cells in the triangulation."""
        return self.T.num_finite_cells()
    @_dependent_property
    def num_infinite_verts(self):
        r"""int: The number of infinite vertices in the triangulation."""
        return self.T.num_infinite_verts()
    @_dependent_property
    def num_infinite_cells(self):
        r"""int: The number of infinite cells in the triangulation."""
        return self.T.num_infinite_cells()
    @_dependent_property
    def num_infinite_edges(self):
        r"""int: The number of infinite edges in the triangulation."""
        return self.T.num_infinite_faces(1)
    @_dependent_property
    def num_infinite_facets(self):
        r"""int: The number of infinite facets in the triangulation."""
        return self.T.num_infinite_faces(D-1)
    @_dependent_property
    def num_verts(self): 
        r"""int: The total number of vertices (finite + infinite) in the 
        triangulation."""
        return self.T.num_verts()
    @_dependent_property
    def num_cells(self):
        r"""int: The total number of cells (finite + infinite) in the 
        triangulation."""
        return self.T.num_cells()
    @_dependent_property
    def num_edges(self):
        r"""int: The total number of edges (finite + infinite) in the 
        triangulation."""
        return self.T.num_faces(1)
    @_dependent_property
    def num_facets(self):
        r"""int: The total number of facets (finite + infinite) in the 
        triangulation."""
        return self.T.num_faces(D - 1)

    def num_finite_faces(self, int d):
        r"""The number of finite faces of dimension `d` in the triangulation.

        Args:
            d (int): Number of dimensions.

        Returns:
            int: The number of finite faces with dimensionality `d` in the 
                triangulation.

        """
        return self.T.num_finite_faces(d)

    def num_infinite_faces(self, int d):
        r"""The number of infinite faces of dimension `d` in the triangulation.

        Args:
            d (int): Number of dimensions.

        Returns:
            int: The number of infinite faces with dimensionality `d` in the 
                triangulation.

        """
        return self.T.num_infinite_faces(d)

    def num_faces(self, int d):
        r"""The total number of faces of dimension `d` in the triangulation.

        Args:
            d (int): Number of dimensions.

        Returns:
            int: The number of faces with dimensionality `d` in the 
                triangulation.

        """
        return self.T.num_faces(d)

    @_dependent_property
    def infinite_vertex(self):
        r"""DelaunayD_vertex: The infinite vertex."""
        cdef Delaunay_with_info_D[info_t].Vertex x = self.T.infinite_vertex()
        cdef DelaunayD_vertex out = DelaunayD_vertex()
        out.assign(self.T, x)
        return out

    @_update_to_tess
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def insert(self, np.ndarray[double, ndim=2, mode="c"] pts not None):
        r"""Insert points into the triangulation.

        Args:
            pts (:obj:`ndarray` of :obj:`float64`): Array of 3D cartesian 
                points to insert into the triangulation. 

        """
        global np_info, np_info_t
        if pts.shape[0] == 0:
            return
        cdef int Nold, Nnew, m
        Nold = self.n
        Nnew, m = pts.shape[0], pts.shape[1]
        if Nnew == 0:
            return
        assert(m == D)
        cdef np.ndarray[np_info_t, ndim=1] idx
        idx = np.arange(Nold, Nold+Nnew).astype(np_info)
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            self.T.insert(&pts[0,0], &idx[0], <info_t>Nnew)
        self.n += Nnew
        self.n_per_insert.append(Nnew)

    @_update_to_tess
    def clear(self):
        r"""Removes all vertices and cells from the triangulation."""
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            self.T.clear()

    @_dependent_property
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def vertices(self):
        r"""ndarray: The x,y,z coordinates of the vertices"""
        cdef np.ndarray[np.float64_t, ndim=2] out
        out = np.zeros([self.n, self.num_dims], 'float64')
        if self.n == 0:
            return out
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            self.T.info_ordered_vertices(&out[0,0])
        return out

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def voronoi_volumes(self):
        r"""np.ndarray of float64: Array of voronoi cell volumes for vertices in 
        the triangulation. The volumes are in the order in which the vertices 
        were added to the triangulation."""
        cdef np.ndarray[np.float64_t, ndim=1] out
        out = np.empty(self.num_finite_verts, 'float64')
        if self.n == 0:
            return out
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            self.T.dual_volumes(&out[0])
        return out

    @_update_to_tess
    def remove(self, DelaunayD_vertex x):
        r"""Remove a vertex from the triangulation. 

        Args: 
            x (DelaunayD_vertex): Vertex that should be removed. 

        """
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            self.T.remove(x.x)

    def get_vertex(self, np_info_t index):
        r"""Get the vertex object corresponding to the given index. 

        Args: 
            index (np_info_t): Index of vertex that should be found. 

        Returns: 
            DelaunayD_vertex: Vertex corresponding to the given index. If the 
                index is not found, the infinite vertex is returned. 

        """
        cdef Delaunay_with_info_D[info_t].Vertex v
        with nogil, cython.boundscheck(False), cython.wraparound(False):
            v = self.T.get_vertex(index)
        cdef DelaunayD_vertex out = DelaunayD_vertex()
        out.assign(self.T, v)
        return out

    def locate(self, np.ndarray[np.float64_t, ndim=1] pos,
               DelaunayD_cell start = None):
        r"""Get the vertex/cell/facet/edge that a given point is a part of.

        Args:
            pos (:obj:`ndarray` of float64): x,y coordinates.   
            start (DelaunayD_cell, optional): Cell to start the search at. 

        Returns:

        """
        assert(len(pos) == self.num_dims)
        cdef int lt
        lt = 999
        cdef DelaunayD_face f = DelaunayD_face()
        cdef DelaunayD_facet ft = DelaunayD_facet()
        cdef DelaunayD_cell c = DelaunayD_cell()
        cdef int ind
        if start is not None:
            c.assign(self.T, self.T.locate(&pos[0], lt, f.x, ft.x, start.x))
        else:
            c.assign(self.T, self.T.locate(&pos[0], lt, f.x, ft.x))
        print(lt)
        assert(lt != 999)
        if lt == 0: # vertex
            return c.vertex(f.index(0))
        elif lt == 1: # face
            return f
        elif lt == 2: # facet
            ind = ft.ind
            ft = c.facet(ind)
            return ft
        elif lt == 3: # cell
            return c
        elif lt == 4:
            print("Point {} is outside the convex hull.".format(pos))
            return c
        elif lt == 5:
            print("Point {} is outside the affine hull.".format(pos))
            return 0
        else:
            raise RuntimeError("Value of {} not expected from CGAL locate.".format(lt))

    @property
    def all_verts_begin(self):
        r"""DelaunayD_vertex_iter: Starting vertex for all vertices in the 
        triangulation."""
        return DelaunayD_vertex_iter(self, 'all_begin')
    @property
    def all_verts_end(self):
        r"""DelaunayD_vertex_iter: Final vertex for all vertices in the 
        triangulation."""
        return DelaunayD_vertex_iter(self, 'all_end')
    @property
    def all_verts(self):
        r"""DelaunayD_vertex_range: Iterable for all vertices in the 
        triangulation."""
        return DelaunayD_vertex_range(self.all_verts_begin, 
                                      self.all_verts_end)
    @property
    def finite_verts(self):
        r"""DelaunayD_vertex_range: Iterable for finite vertices in the 
        triangulation."""
        return DelaunayD_vertex_range(self.all_verts_begin, 
                                      self.all_verts_end, finite = True)

    @property
    def all_cells_begin(self):
        r"""DelaunayD_cell_iter: Starting cell for all cells in the triangulation."""
        return DelaunayD_cell_iter(self, 'all_begin')
    @property
    def all_cells_end(self):
        r"""DelaunayD_cell_iter: Finall cell for all cells in the triangulation."""
        return DelaunayD_cell_iter(self, 'all_end')
    @property
    def all_cells(self):
        r"""DelaunayD_cell_range: Iterable for all cells in the
        triangulation."""
        return DelaunayD_cell_range(self.all_cells_begin,
                                    self.all_cells_end)
    @property
    def finite_cells(self):
        r"""DelaunayD_cell_range: Iterable for finite cells in the
        triangulation."""
        return DelaunayD_cell_range(self.all_cells_begin,
                                    self.all_cells_end, finite = True)

    def mirror_index(self, DelaunayD_cell x, int i):
        r"""Get the index of a cell with respect to its ith neighbor. 

        Args: 
            x (DelaunayD_cell): Cell to get mirrored index for. 
            i (int): Index of neighbor that should be used to determine the 
                mirrored index. 

        Returns: 
            int: Index of cell x with respect to its ith neighbor. 

        """
        return x.mirror_index(i)

    def mirror_vertex(self, DelaunayD_cell x, int i):
        r"""Get the vertex of a cell's ith neighbor opposite to the cell. 

        Args: 
            x (DelaunayD_cell): Cell. 
            i (int): Index of neighbor that should be used to determine the 
                mirrored vertex. 

        Returns:
            DelaunayD_vertex: Vertex in the ith neighboring cell of cell x, 
                that is opposite to cell x. 

        """
        return x.mirror_vertex(i)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def outgoing_points(self,
                        np.ndarray[np.float64_t, ndim=2] left_edges,
                        np.ndarray[np.float64_t, ndim=2] right_edges):
        r"""Get the indices of points in tets that intersect a set of boxes.

        Args: 
            left_edges (np.ndarray of float64): (m, n) array of m box mins in n 
                dimensions. 
            right_edges (np.ndarray of float64): (m, n) array of m box maxs in n 
                dimensions. 

        Returns: 

        """
        assert(left_edges.shape[1] == self.num_dims)
        assert(left_edges.shape[0] == right_edges.shape[0])
        assert(left_edges.shape[1] == right_edges.shape[1])
        cdef uint64_t nbox = <uint64_t>left_edges.shape[0]
        cdef vector[vector[info_t]] vout
        if (nbox > 0):
            with nogil, cython.boundscheck(False), cython.wraparound(False):
                vout = self.T.outgoing_points(nbox,
                                              &left_edges[0,0],
                                              &right_edges[0,0])
        assert(vout.size() == nbox)
        # Transfer values to array
        cdef uint64_t i, j
        cdef object out = [None for i in xrange(vout.size())]
        for i in xrange(vout.size()):
            out[i] = np.empty(vout[i].size(), np_info)
            for j in xrange(vout[i].size()):
                out[i][j] = vout[i][j]
        return out




r"""Routines for tracking the scaling of the triangulation routines."""
import numpy as np
import time
import os
import cProfile
import pstats
from cgal4py import parallel, delaunay
from test_cgal4py import run_test
import matplotlib.pyplot as plt
np.random.seed(10)


def stats_run(func, npart, nproc, ndim, periodic=False, use_mpi=False,
              use_python=False, use_buffer=False, overwrite=False,
              display=False, suppress_final_output=False):
    r"""Get timing stats using :package:`cProfile`.

    Args:
        func (str): Name of the function that should be run. Values include:
            'Delaunay': Full triangulation.
            'VoronoiVolumes': Cell volumes from triangulation.
        npart (int): Number of particles.
        nproc (int): Number of processors.
        ndim (int): Number of dimensions.
        periodic (bool, optional): If True, the domain is assumed to be
            periodic. Defaults to False.
        use_mpi (bool, optional): If True, the MPI parallelized version is used
            instead of the version using the multiprocessing package. Defaults
            to False.
        use_python (bool, optional): If True and `use_mpi == True`, then MPI
            communications are done in python using `mpi4py`. Defaults to
            False.
        use_buffer (bool, optional): If True and `use_mpi == True`, then buffer
            versions of MPI communications will be used rather than indirect
            communications of python objects via pickling. Defaults to False.
        suppress_final_output (bool, optional): If True, the final output 
            from spawned MPI processes is suppressed. This is mainly for
            timing purposes. Defaults to False.
        overwrite (bool, optional): If True, the existing file for this
            set of input parameters if overwritten. Defaults to False.
        display (bool, optional): If True, display the profile results.
            Defaults to False.

    """
    perstr = ""
    if periodic:
        perstr = "_periodic"
    mpistr = ""
    bufstr = ""
    pycstr = ""
    outstr = ""
    if use_mpi:
        mpistr = "_mpi"
        if suppress_final_output:
            outstr = "_noout"
        if use_python:
            pycstr = "_python"
            if use_buffer:
                bufstr = "_buffer"
    fname_stat = 'stat_{}_{}part_{}proc_{}dim{}{}{}{}{}.txt'.format(
        func, npart, nproc, ndim, perstr, mpistr, pycstr, bufstr, outstr)
    if overwrite or not os.path.isfile(fname_stat):
        cProfile.run(
            "from cgal4py.tests.test_cgal4py import run_test; "+
            "run_test({}, {}, nproc={}, func_name='{}',".format(
                npart, ndim, nproc, func) +
            "periodic={},use_mpi={},use_python={},use_buffer={},".format(
                periodic, use_mpi, use_python, use_buffer) +
            "suppress_final_output={})".format(suppress_final_output),
            fname_stat)
    if display:
        p = pstats.Stats(fname_stat)
        p.sort_stats('time').print_stats(10)
        return p
    return fname_stat


def time_run(func, npart, nproc, ndim, nrep=1, periodic=False, use_mpi=False,
             use_python=False, use_buffer=False, suppress_final_output=False):
    r"""Get runing times using :package:`time`.

    Args:
        func (str): Name of the function that should be run. Values include:
            'Delaunay': Full triangulation.
            'VoronoiVolumes': Cell volumes from triangulation.
        npart (int): Number of particles.
        nproc (int): Number of processors.
        ndim (int): Number of dimensions.
        nrep (int, optional): Number of times the run should be performed to
            get an average. Defaults to 1.
        periodic (bool, optional): If True, the domain is assumed to be
            periodic. Defaults to False.
        use_mpi (bool, optional): If True, the MPI parallelized version is used
            instead of the version using the multiprocessing package. Defaults
            to False.
        use_python (bool, optional): If True and `use_mpi == True`, then MPI
            communications are done in python using `mpi4py`. Defaults to
            False.
        use_buffer (bool, optional): If True and `use_mpi == True`, then buffer
            versions of MPI communications will be used rather than indirect
            communications of python objects via pickling. Defaults to False.
        suppress_final_output (bool, optional): If True, the final output 
            from spawned MPI processes is suppressed. This is mainly for
            timing purposes. Defaults to False.

    """
    times = np.empty(nrep, 'float')
    for i in range(nrep):
        t1 = time.time()
        run_test(npart, ndim, nproc=nproc, func_name=func,
                 periodic=periodic, use_mpi=use_mpi,
                 use_python=use_python, use_buffer=use_buffer,
                 suppress_final_output=suppress_final_output)
        t2 = time.time()
        times[i] = t2 - t1
    return np.mean(times), np.std(times)


def strong_scaling(func, npart=1e6, nrep=1, periodic=False, use_mpi=False,
                   use_python=False, use_buffer=False, overwrite=True,
                   suppress_final_output=False):
    r"""Plot the scaling with number of processors for a particular function.

    Args:
        func (str): Name of the function that should be run. Values include:
            'Delaunay': Full triangulation.
            'VoronoiVolumes': Cell volumes from triangulation.
        npart (int, optional): Number of particles. Defaults to 1e6.
        nrep (int, optional): Number of times the run should be performed to
            get an average. Defaults to 1.
        periodic (bool, optional): If True, the domain is assumed to be
            periodic. Defaults to False.
        use_mpi (bool, optional): If True, the MPI parallelized version is used
            instead of the version using the multiprocessing package. Defaults
            to False.
        use_python (bool, optional): If True and `use_mpi == True`, then MPI
            communications are done in python using `mpi4py`. Defaults to
            False.
        use_buffer (bool, optional): If True and `use_mpi == True`, then buffer
            versions of MPI communications will be used rather than indirect
            communications of python objects via pickling. Defaults to False.
        suppress_final_output (bool, optional): If True, the final output 
            from spawned MPI processes is suppressed. This is mainly for
            timing purposes. Defaults to False.
        overwrite (bool, optional): If True, the existing file for this
            set of input parameters if overwritten. Defaults to False.

    """
    npart = int(npart)
    perstr = ""
    if periodic:
        perstr = "_periodic"
    mpistr = ""
    pycstr = ""
    bufstr = ""
    outstr = ""
    if use_mpi:
        mpistr = "_mpi"
        if suppress_final_output:
            outstr = "_noout"
        if use_python:
            pycstr = "_python"
            if use_buffer:
                bufstr = "_buffer"
    fname_plot = 'plot_strong_scaling_{}_nproc_{}part{}{}{}{}{}.png'.format(
        func, npart, perstr, mpistr, pycstr, bufstr, outstr)
    nproc_list = [1, 2, 4, 8, 16]
    ndim_list = [2, 3, 4]
    clr_list = ['b', 'r', 'g', 'm']
    times = np.empty((len(nproc_list), len(ndim_list), 2), 'float')
    for j, nproc in enumerate(nproc_list):
        for i, ndim in enumerate(ndim_list):
            times[j, i, 0], times[j, i, 1] = time_run(
                func, npart, nproc, ndim, nrep=nrep,
                periodic=periodic, use_mpi=use_mpi,
                use_python=use_python, use_buffer=use_buffer,
                suppress_final_output=suppress_final_output)
            print("Finished {}D on {}.".format(ndim, nproc))
    fig, axs = plt.subplots(1, 1)
    for i in range(len(ndim_list)):
        ndim = ndim_list[i]
        clr = clr_list[i]
        axs.errorbar(nproc_list, times[:, i, 0], yerr=times[:, i, 1],
                     fmt=clr, label='ndim = {}'.format(ndim))
    axs.set_xlabel("# of Processors")
    axs.set_ylabel("Time (s)")
    axs.legend()
    fig.savefig(fname_plot)
    print('    '+fname_plot)


def weak_scaling(func, npart=1e4, nrep=1, periodic=False, use_mpi=False,
                 use_python=False, use_buffer=False, overwrite=True,
                 suppress_final_output=False):
    r"""Plot the scaling with number of processors with a constant number of
    particles per processor for a particular function.

    Args:
        func (str): Name of the function that should be run. Values include:
            'Delaunay': Full triangulation.
            'VoronoiVolumes': Cell volumes from triangulation.
        npart (int, optional): Number of particles per processor. Defaults to
            1e4.
        nrep (int, optional): Number of times the run should be performed to
            get an average. Defaults to 1.
        periodic (bool, optional): If True, the domain is assumed to be
            periodic. Defaults to False.
        use_mpi (bool, optional): If True, the MPI parallelized version is used
            instead of the version using the multiprocessing package. Defaults
            to False.
        use_python (bool, optional): If True and `use_mpi == True`, then MPI
            communications are done in python using `mpi4py`. Defaults to
            False.
        use_buffer (bool, optional): If True and `use_mpi == True`, then buffer
            versions of MPI communications will be used rather than indirect
            communications of python objects via pickling. Defaults to False.
        suppress_final_output (bool, optional): If True, the final output 
            from spawned MPI processes is suppressed. This is mainly for
            timing purposes. Defaults to False.
        overwrite (bool, optional): If True, the existing file for this
            set of input parameters if overwritten. Defaults to False.

    """
    npart = int(npart)
    perstr = ""
    if periodic:
        perstr = "_periodic"
    mpistr = ""
    pycstr = ""
    bufstr = ""
    outstr = ""
    if use_mpi:
        mpistr = "_mpi"
        if suppress_final_output:
            outstr = "_noout"
        if use_python:
            pycstr = "_python"
            if use_buffer:
                bufstr = "_buffer"
    fname_plot = 'plot_weak_scaling_{}_nproc_{}part{}{}{}{}{}.png'.format(
        func, npart, perstr, mpistr, pycstr, bufstr, outstr)
    nproc_list = [1, 2, 4, 8, 16]
    ndim_list = [2, 3]
    clr_list = ['b', 'r', 'g', 'm']
    times = np.empty((len(nproc_list), len(ndim_list), 2), 'float')
    for j, nproc in enumerate(nproc_list):
        for i, ndim in enumerate(ndim_list):
            times[j, i, 0], times[j, i, 1] = time_run(
                func, npart*nproc, nproc, ndim, nrep=nrep,
                periodic=periodic, use_mpi=use_mpi,
                use_python=use_python, use_buffer=use_buffer,
                suppress_final_output=suppress_final_output)
    fig, axs = plt.subplots(1, 1)
    for i in range(len(ndim_list)):
        ndim = ndim_list[i]
        clr = clr_list[i]
        axs.errorbar(nproc_list, times[:, i, 0], yerr=times[:, i, 1],
                     fmt=clr, label='ndim = {}'.format(ndim))
    axs.set_xlabel("# of Processors")
    axs.set_ylabel("Time (s)")
    axs.legend()
    fig.savefig(fname_plot)
    print('    '+fname_plot)



cimport numpy as np
from libcpp.vector cimport vector
from libcpp.set cimport set as cset
from libcpp.pair cimport pair
from libcpp cimport bool
from libc.stdint cimport uint32_t, uint64_t, int32_t, int64_t

cdef extern from "c_delaunay3.hpp":
    cdef int VALID

    cdef cppclass Delaunay_with_info_3[Info] nogil:
        Delaunay_with_info_3() except +
        Delaunay_with_info_3(double *pts, Info *val, uint32_t n) except +
        bool updated
        bool is_valid() const
        uint32_t num_finite_verts() const
        uint32_t num_finite_edges() const
        uint32_t num_finite_facets() const
        uint32_t num_finite_cells() const
        uint32_t num_infinite_verts() const
        uint32_t num_infinite_edges() const
        uint32_t num_infinite_facets() const
        uint32_t num_infinite_cells() const
        uint32_t num_verts() const
        uint32_t num_edges() const
        uint32_t num_facets() const
        uint32_t num_cells() const

        bool is_equal(const Delaunay_with_info_3[Info] other) const

        cppclass Vertex
        cppclass Edge
        cppclass Facet
        cppclass Cell

        Vertex infinite_vertex() const

        void insert(double *, Info *val, uint32_t n) except +
        void remove(Vertex) except +
        void clear() except + 
        Vertex move(Vertex v, double *pos) except + 
        Vertex move_if_no_collision(Vertex v, double *pos) except +

        void write_to_file(const char* filename) except +
        void read_from_file(const char* filename) except +
        I serialize[I](I &n, I &m, int32_t &d,
                       double* vert_pos, Info* vert_info,
                       I* cells, I* neighbors) const
        Info serialize_idxinfo[I](I &n, I &m, int32_t &d,
                                  Info* cells, I* neighbors) const
        I serialize_info2idx[I](I &n, I &m, int32_t &d,
                                I* faces, I* neighbors,
                                Info max_info, I* idx) const
        void deserialize[I](I n, I m, int32_t d,
                         double* vert_pos, Info* vert_info,
                         I* cells, I* neighbors, I idx_inf)
        void deserialize_idxinfo[I](I n, I m, int32_t d, double* vert_pos,
                                    I* cells, I* neighbors, I idx_inf)

        Vertex get_vertex(Info index) except +
        Cell locate(double* pos, int& lt, int& li, int& lj)
        Cell locate(double* pos, int& lt, int& li, int& lj, Cell c)

        void info_ordered_vertices(double* pos)
        void vertex_info(Info* verts)
        void edge_info(Info* edges)

        cppclass All_verts_iter:
            All_verts_iter()
            All_verts_iter& operator++()
            All_verts_iter& operator--()
            bool operator==(All_verts_iter other)
            bool operator!=(All_verts_iter other)
            Vertex vertex()
        All_verts_iter all_verts_begin()
        All_verts_iter all_verts_end()

        cppclass Vertex:
            Vertex()
            Vertex(All_verts_iter v)
            bool operator==(Vertex other)
            bool operator!=(Vertex other)
            void point(double* out)
            vector[double] point()
            Info info()
            Cell cell()
            void set_point(double* x)
            void set_cell(Cell c)

        cppclass All_edges_iter:
            All_edges_iter()
            All_edges_iter& operator++()
            All_edges_iter& operator--()
            bool operator==(All_edges_iter other)
            bool operator!=(All_edges_iter other)
        All_edges_iter all_edges_begin()
        All_edges_iter all_edges_end()

        cppclass Edge:
            Edge()
            Edge(All_edges_iter it)
            Edge(Cell x, int i1, int i2)
            bool operator==(Edge other)
            bool operator!=(Edge other)
            Cell cell()
            int ind1()
            int ind2()
            Vertex vertex(int i)
            Vertex v1()
            Vertex v2()

        cppclass All_facets_iter:
            All_facets_iter()
            All_facets_iter& operator++()
            All_facets_iter& operator--()
            bool operator==(All_facets_iter other)
            bool operator!=(All_facets_iter other)
        All_facets_iter all_facets_begin()
        All_facets_iter all_facets_end()

        cppclass Facet:
            Facet()
            Facet(All_facets_iter it)
            Facet(Cell x, int i1)
            bool operator==(Facet other)
            bool operator!=(Facet other)
            Cell cell()
            int ind()
            Vertex vertex(int i)
            Edge edge(int i)

        cppclass All_cells_iter:
            All_cells_iter()
            All_cells_iter& operator++()
            All_cells_iter& operator--()
            bool operator==(All_cells_iter other)
            bool operator!=(All_cells_iter other)
        All_cells_iter all_cells_begin()
        All_cells_iter all_cells_end()

        cppclass Cell:
            Cell()
            Cell(All_cells_iter c)
            Cell(Vertex v1, Vertex v2, Vertex v3, Vertex v4)
            Cell(Vertex v1, Vertex v2, Vertex v3, Vertex v4,
                 Cell n1, Cell n2, Cell n3, Cell n4)
            bool operator==(Cell other)
            bool operator!=(Cell other)

            Facet facet(int i)

            Vertex vertex(int i)
            bool has_vertex(Vertex v)
            bool has_vertex(Vertex v, int *i)
            int ind(Vertex v)

            Cell neighbor(int i)
            bool has_neighbor(Cell c)
            bool has_neighbor(Cell c, int *i)
            int ind(Cell c)

            void set_vertex(int i, Vertex v)
            void set_vertices()
            void set_vertices(Vertex v1, Vertex v2, Vertex v3, Vertex v4)
            void set_neighbor(int i, Cell c)
            void set_neighbors()
            void set_neighbors(Cell c1, Cell c2, Cell c3, Cell c4)

        bool are_equal(const Edge e1, const Edge e2) 
        bool are_equal(const Facet f1, const Facet f2) 

        bool is_infinite(Vertex x)
        bool is_infinite(Edge x)
        bool is_infinite(Facet x)
        bool is_infinite(Cell x)
        bool is_infinite(All_verts_iter x)
        bool is_infinite(All_edges_iter x)
        bool is_infinite(All_facets_iter x)
        bool is_infinite(All_cells_iter x)

        bool is_edge(Vertex x1, Vertex x2, Cell& c, int& i, int& j) const
        bool is_facet(Vertex x1, Vertex x2, Vertex x3, Cell& c, int& i, int& j, int& k) const
        bool is_cell(Vertex x1, Vertex x2, Vertex x3, Vertex x4,
                     Cell& c, int& i1, int& i2, int& i3, int& i4) const

        vector[Vertex] incident_vertices(Vertex x)
        vector[Edge] incident_edges(Vertex x)
        vector[Facet] incident_facets(Vertex x)
        vector[Cell] incident_cells(Vertex x)

        vector[Vertex] incident_vertices(Edge x)
        vector[Edge] incident_edges(Edge x)
        vector[Facet] incident_facets(Edge x)
        vector[Cell] incident_cells(Edge x)

        vector[Vertex] incident_vertices(Facet x)
        vector[Edge] incident_edges(Facet x)
        vector[Facet] incident_facets(Facet x)
        vector[Cell] incident_cells(Facet x)

        vector[Vertex] incident_vertices(Cell x)
        vector[Edge] incident_edges(Cell x)
        vector[Facet] incident_facets(Cell x)
        vector[Cell] incident_cells(Cell x)

        Vertex nearest_vertex(double* pos)

        Facet mirror_facet(Facet x) const
        int mirror_index(Cell x, int i) const
        Vertex mirror_vertex(Cell x, int i) const

        void circumcenter(Cell x, double* out) const
        double dual_volume(const Vertex v) const
        void dual_volumes(double* vols) const
        double length(const Edge e) const

        bool flip(Cell x, int i, int j)
        bool flip(Edge x)
        bool flip(Cell x, int i)
        bool flip(Facet x)
        void flip_flippable(Cell x, int i, int j)
        void flip_flippable(Edge x)
        void flip_flippable(Cell x, int i)
        void flip_flippable(Facet x)

        pair[vector[Cell],vector[Facet]] find_conflicts(double* pos, Cell start)

        int side_of_cell(const double* pos, Cell c, int& lt, int& li, int& lj) const
        int side_of_edge(const double* pos, const Edge e, int& lt, int& li) const
        int side_of_facet(const double* pos, const Facet f, int& lt, int& li, int& lj) const 
        # int side_of_circle(const Facet f, const double* pos)
        int side_of_sphere(const Cell c, const double* pos)
        bool is_Gabriel(const Edge e)
        bool is_Gabriel(const Facet f)

        vector[vector[Info]] outgoing_points(uint64_t nbox,
                                             double *left_edges, double *right_edges)
        void boundary_points(double *left_edge, double *right_edge, bool periodic,
                             vector[Info]& lx, vector[Info]& ly, vector[Info]& lz,
                             vector[Info]& rx, vector[Info]& ry, vector[Info]& rz,
                             vector[Info]& alln)



import test_cgal4py
import test_delaunay2
import test_delaunay3
import test_plot
import test_utils
import scaling


__all__ = ["test_cgal4py", "test_delaunay2", "test_delaunay3", "test_plot",
           "test_utils", "scaling"]



r"""Tests for domain decomposition methods."""
import numpy as np
from nose.tools import assert_raises
from cgal4py import domain_decomp


N = 100
leafsize = 10
pts2 = np.random.rand(N, 2).astype('float64')
left_edge2 = np.zeros(2, 'float64')
right_edge2 = np.ones(2, 'float64')
pts3 = np.random.rand(N, 3).astype('float64')
left_edge3 = np.zeros(3, 'float64')
right_edge3 = np.ones(3, 'float64')


def test_tree():
    tree2 = domain_decomp.tree('kdtree', pts2, left_edge2, right_edge2,
                               periodic=False, leafsize=leafsize)
    tree3 = domain_decomp.tree('kdtree', pts3, left_edge3, right_edge3,
                               periodic=False, leafsize=leafsize)
    assert_raises(ValueError, domain_decomp.tree, 'invalid',
                  pts2, left_edge2, right_edge2, False)
    del tree2, tree3


def test_GenericLeaf():
    leaf2 = domain_decomp.GenericLeaf(N, left_edge2, right_edge2)
    leaf3 = domain_decomp.GenericLeaf(N, left_edge3, right_edge3)
    del leaf2, leaf3


def test_process_leaves():
    # With information
    tree2 = domain_decomp.tree('kdtree', pts2, left_edge2, right_edge2,
                               periodic=False, leafsize=leafsize)
    tree3 = domain_decomp.tree('kdtree', pts3, left_edge3, right_edge3,
                               periodic=False, leafsize=leafsize)
    leaves2 = domain_decomp.process_leaves(tree2.leaves, left_edge2,
                                           right_edge2, False)
    leaves3 = domain_decomp.process_leaves(tree3.leaves, left_edge3,
                                           right_edge3, False)
    # Without information
    left_edges = np.array([[0.0, 0.0],
                           [0.0, 0.5],
                           [0.5, 0.0],
                           [0.5, 0.5]], 'float')
    right_edges = np.array([[0.5, 0.5],
                            [0.5, 1.0],
                            [1.0, 0.5],
                            [1.0, 1.0]], 'float')
    leaves = [domain_decomp.GenericLeaf(
        5, left_edges[i, :], right_edges[i, :]) for i in range(2)]
    leaves = domain_decomp.process_leaves(leaves, left_edge2,
                                          right_edge2, False)
    del leaves2, leaves3


def test_GenericTree():
    # With information
    tree2 = domain_decomp.tree('kdtree', pts2, left_edge2, right_edge2,
                               periodic=False, leafsize=leafsize)
    tree3 = domain_decomp.tree('kdtree', pts3, left_edge3, right_edge3,
                               periodic=False, leafsize=leafsize)
    leaves2 = domain_decomp.process_leaves(tree2.leaves, left_edge2,
                                           right_edge2, False)
    leaves3 = domain_decomp.process_leaves(tree3.leaves, left_edge3,
                                           right_edge3, False)
    # Without information
    left_edges = np.array([[0.0, 0.0],
                           [0.0, 0.5],
                           [0.5, 0.0],
                           [0.5, 0.5]], 'float')
    right_edges = np.array([[0.5, 0.5],
                            [0.5, 1.0],
                            [1.0, 0.5],
                            [1.0, 1.0]], 'float')
    leaves = [domain_decomp.GenericLeaf(
        5, left_edges[i, :], right_edges[i, :]) for i in range(2)]
    tree = domain_decomp.GenericTree(
        np.arange(5*left_edges.shape[0]).astype('int'),
        leaves, left_edge2, right_edge2, False)
    del tree, leaves2, leaves3



cimport numpy as np
from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libcpp cimport bool
from libc.stdint cimport uint32_t, uint64_t, int64_t, int32_t

cdef extern from "c_tools.hpp":
    bool intersect_sph_box(uint32_t ndim, double *c, double r, double *le, double *re) nogil
    bool arg_tLT[I](I *cells, uint32_t *idx_verts, uint32_t ndim, uint64_t i1, uint64_t i2) nogil
    bool tEQ[I](I *cells, uint32_t ndim, int64_t i1, int64_t i2) nogil
    bool tGT[I](I *cells, uint32_t ndim, int64_t i1, int64_t i2) nogil
    bool tLT[I](I *cells, uint32_t ndim, int64_t i1, int64_t i2) nogil
    int64_t arg_partition[I](I *arr, I *idx, uint32_t ndim,
                             int64_t l, int64_t r, int64_t p) nogil
    void arg_quickSort[I](I *arr, I *idx, uint32_t ndim,
                          int64_t l, int64_t r) nogil
    int64_t partition_tess[I](I *cells, I *neigh, I *idx, uint32_t ndim,
                              int64_t l, int64_t r, int64_t p) nogil
    void quickSort_tess[I](I *cells, I *neigh, I *idx, uint32_t ndim,
                           int64_t l, int64_t r) nogil
    void sortCellVerts[I](I *cells, I *neigh, uint64_t ncells, uint32_t ndim) nogil
    void sortSerializedTess[I](I *cells, I *neigh, 
                               uint64_t ncells, uint32_t ndim) nogil
    int64_t arg_partition_tess[I](I *cells, uint32_t *idx_verts, uint64_t *idx_cells, 
                                  uint32_t ndim, int64_t l, int64_t r, int64_t p) nogil
    void arg_quickSort_tess[I](I *cells, uint32_t *idx_verts, uint64_t *idx_cells, 
                               uint32_t ndim, int64_t l, int64_t r) nogil
    void arg_sortCellVerts[I](I *cells, uint32_t *idx_verts, uint64_t ncells, uint32_t ndim) nogil
    void arg_sortSerializedTess[I](I *cells, uint64_t ncells, uint32_t ndim,
                                   uint32_t *idx_verts, uint64_t *idx_cells) nogil
    void swap_cells[I](I *verts, I *neigh, uint32_t ndim, uint64_t i1, uint64_t i2) nogil
    cdef cppclass SerializedLeaf[I] nogil:
        SerializedLeaf() except +
        SerializedLeaf(int _id, uint32_t _ndim, int64_t _ncells, I _idx_inf,
                       I *_verts, I *_neigh,
                       uint32_t *_sort_verts, uint64_t *_sort_cells,
                       uint64_t idx_start, uint64_t idx_stop) except +
        int id
        uint32_t ndim
        int64_t ncells
        uint64_t idx_start
        uint64_t idx_stop
        I idx_inf
        I *verts
        I *neigh
        uint32_t *sort_verts
        uint64_t *sort_cells
        bool init_from_file
        void write_to_file(const char* filename)
        int64_t read_from_file(const char* filename)
        void cleanup()
    cdef cppclass ConsolidatedLeaves[I] nogil:
        ConsolidatedLeaves() except +
        ConsolidatedLeaves(uint32_t _ndim, I _idx_inf, int64_t _max_ncells,
                           I *_verts, I *_neigh) except +
        ConsolidatedLeaves(uint32_t _ndim, int64_t _ncells, I _idx_inf,
                           int64_t _max_ncells, I *_verts, I *_neigh) except +
        ConsolidatedLeaves(uint32_t _ndim, int64_t _ncells, I _idx_inf,
                           int64_t _max_ncells, I *_verts, I *_neigh,
                           uint64_t n_split_map, I *key_split_map, uint64_t *val_split_map,
                           uint64_t n_inf_map, I *key_inf_map, uint64_t *val_inf_map) except +
        int64_t ncells
        int64_t max_ncells
        I *allverts
        I *allneigh
        uint64_t size_split_map()
        uint64_t size_inf_map()
        void get_split_map(I *keys, uint64_t *vals) 
        void get_inf_map(I *keys, uint64_t *vals)
        void cleanup()
        void add_leaf[leafI](SerializedLeaf[leafI] leaf)
        void add_leaf_fromfile(const char *filename)
        int64_t count_inf()
        void add_inf()

ctypedef SerializedLeaf[uint32_t] sLeaf32
ctypedef SerializedLeaf[uint64_t] sLeaf64
ctypedef vector[sLeaf32] sLeaves32
ctypedef vector[sLeaf64] sLeaves64



r"""Tests for the 2D Periodic Delaunay Extension.

.. todo:
   * value test circumcenter
   * test includes_edge

"""
import numpy as np
import os
from cgal4py.delaunay import PeriodicDelaunay2 as Delaunay2

left_edge = -2*np.ones(2, 'float64')
right_edge = 2*np.ones(2, 'float64')
pts = np.array([[-0.49419885869540180, -0.07594397977563715],
                [-0.06448037284989526,  0.49582484963658130],
                [+0.49111543670946320,  0.09383830681375946],
                [-0.34835358086909700, -0.35867782576523670],
                [-1,     -1],
                [-1,      1],
                [+1,     -1],
                [+1,      1]], 'float64')
pts_dup = np.concatenate([pts, np.reshape(pts[0, :], (1, pts.shape[1]))])
nsheets = np.array([3, 3], 'int32')
nverts_fin = pts.shape[0]
nverts_inf = 0
nverts = nverts_fin + nverts_inf
nedges_fin = 24  # 17 in non-periodic
nedges_inf = 0   # 4 in non-periodic
nedges = nedges_fin + nedges_inf
ncells_fin = 16  # 10 in non-periodic
ncells_inf = 0   # 4 in non-periodic
ncells = ncells_fin + ncells_inf


def test_create():
    T = Delaunay2(left_edge, right_edge)
    del T


def test_insert():
    # without duplicates
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    assert(T.is_valid())
    # with duplicates
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts_dup)
    assert(T.is_valid())


def test_equal():
    T1 = Delaunay2(left_edge, right_edge)
    T1.insert(pts)
    T2 = Delaunay2(left_edge, right_edge)
    T2.insert(pts)
    assert(T1.is_equivalent(T2))


def test_num_sheets():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    print(T.num_sheets, T.num_sheets_total)
    assert(np.allclose(T.num_sheets, nsheets))
    assert(T.num_sheets_total == nsheets.prod())


def test_num_verts():
    # without duplicates
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    print(T.num_finite_verts, T.num_infinite_verts, T.num_verts)
    assert(T.num_finite_verts == nverts_fin)
    assert(T.num_infinite_verts == nverts_inf)
    assert(T.num_verts == nverts)
    assert(T.num_stored_verts == T.num_sheets_total*nverts)
    # with duplicates
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts_dup)
    print(T.num_finite_verts, T.num_infinite_verts, T.num_verts)
    assert(T.num_finite_verts == nverts_fin)
    assert(T.num_infinite_verts == nverts_inf)
    assert(T.num_verts == nverts)
    assert(T.num_stored_verts == T.num_sheets_total*nverts)


def test_num_edges():
    # without duplicates
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    print(T.num_finite_edges, T.num_infinite_edges, T.num_edges)
    assert(T.num_finite_edges == nedges_fin)
    assert(T.num_infinite_edges == nedges_inf)
    assert(T.num_edges == nedges)
    assert(T.num_stored_edges == T.num_sheets_total*nedges)
    # with duplicates
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts_dup)
    print(T.num_finite_edges, T.num_infinite_edges, T.num_edges)
    assert(T.num_finite_edges == nedges_fin)
    assert(T.num_infinite_edges == nedges_inf)
    assert(T.num_edges == nedges)
    assert(T.num_stored_edges == T.num_sheets_total*nedges)


def test_num_cells():
    # without duplicates
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    print(T.num_finite_cells, T.num_infinite_cells, T.num_cells)
    assert(T.num_finite_cells == ncells_fin)
    assert(T.num_infinite_cells == ncells_inf)
    assert(T.num_cells == ncells)
    assert(T.num_stored_cells == T.num_sheets_total*ncells)
    # with duplicates
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts_dup)
    print(T.num_finite_cells, T.num_infinite_cells, T.num_cells)
    assert(T.num_finite_cells == ncells_fin)
    assert(T.num_infinite_cells == ncells_inf)
    assert(T.num_cells == ncells)
    assert(T.num_stored_cells == T.num_sheets_total*ncells)


def test_all_verts():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count_fin = count_inf = 0
    for v in T.all_verts:
        print(v.index, v.point)
        if v.is_infinite():
            count_inf += 1
        else:
            count_fin += 1
    count = count_fin + count_inf
    assert(count_fin == T.num_stored_verts)
    assert(count_inf == T.num_infinite_verts)
    assert(count == T.num_stored_verts)


def test_finite_verts():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for v in T.finite_verts:
        assert((not v.is_infinite()))
        count += 1
    assert(count == T.num_stored_verts)


def test_all_edges():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count_fin = count_inf = 0
    for e in T.all_edges:
        if e.is_infinite():
            count_inf += 1
        else:
            count_fin += 1
    count = count_fin + count_inf
    assert(count_fin == T.num_stored_edges)
    assert(count_inf == T.num_infinite_edges)
    assert(count == T.num_stored_edges)


def test_finite_edges():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for e in T.finite_edges:
        assert((not e.is_infinite()))
        count += 1
    print(count)
    assert(count == T.num_stored_edges)


def test_all_cells():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count_fin = count_inf = 0
    for c in T.all_cells:
        if c.is_infinite():
            count_inf += 1
        else:
            count_fin += 1
    count = count_fin + count_inf
    assert(count_fin == T.num_stored_cells)
    assert(count_inf == T.num_infinite_cells)
    assert(count == T.num_stored_cells)


def test_finite_cells():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for c in T.finite_cells:
        assert((not c.is_infinite()))
        count += 1
    assert(count == T.num_stored_cells)


def test_clear():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    T.clear()
    print(T.num_finite_verts, T.num_cells)
    assert(T.num_finite_verts == 0)
    assert(T.num_cells == 0)


def test_io():
    fname = 'test_io2348_2.dat'
    Tout = Delaunay2(left_edge, right_edge)
    Tout.insert(pts)
    Tout.write_to_file(fname)
    Tin = Delaunay2()
    Tin.read_from_file(fname)
    assert(Tout.num_verts == Tin.num_verts)
    assert(Tout.num_cells == Tin.num_cells)
    os.remove(fname)


def test_serialize():
    Tout = Delaunay2(left_edge, right_edge)
    Tout.insert(pts)
    out = Tout.serialize()
    Tin = Delaunay2()
    Tin.deserialize(pts, *out)
    print(Tout.num_cells, Tin.num_cells)
    assert(Tout.num_verts == Tin.num_verts)
    assert(Tout.num_cells == Tin.num_cells)


def test_locate():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    for c in T.finite_cells:
        p = c.center
        print('{}\n{}\n{}'.format(c, p, T.locate(p)))
        assert(c == T.locate(p, c))
        assert(c == T.locate(p))
        assert(c.vertex(0) == T.locate(c.vertex(0).point))
        assert(c.vertex(0) == T.locate(c.vertex(0).point, c))
        # assert(c.edge(0) == T.locate(c.edge(0).midpoint))
        # assert(c.edge(0) == T.locate(c.edge(0).midpoint, c))
        break


def test_get_vertex():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    for i in range(nverts_fin):
        v = T.get_vertex(i)
        assert(np.allclose(v.point, pts[i, :]))


def test_remove():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    v = T.get_vertex(0)
    T.remove(v)
    assert(T.num_verts == (nverts-1))


def test_is_edge():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    assert(T.is_edge(T.get_vertex(0), T.get_vertex(1)))
    assert(not T.is_edge(T.get_vertex(0), T.get_vertex(nverts_fin-1)))


def test_is_cell():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    assert(T.is_cell(T.get_vertex(0), T.get_vertex(1), T.get_vertex(3)))
    assert(not T.is_cell(T.get_vertex(0), T.get_vertex(1),
                         T.get_vertex(nverts_fin-1)))


def test_vert():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    vold = None
    for v in T.all_verts:
        pnt = v.point
        per = v.periodic_point
        idx = v.index
        vol = v.dual_volume
        print(pts[idx, :])
        print(v, idx, vol, pnt, per)
        assert(v == v)
        if vold is not None:
            assert(v != vold)
        if v.is_infinite():
            assert(np.isinf(per).all())
            assert(idx == np.iinfo(np.uint32).max)
            assert(np.isclose(vol, -1))
        else:
            assert(np.allclose(per, pts[idx, :]))
            c = v.cell
            v.set_cell(c)
            v.set_point(per)
        vold = v


def test_edge():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    eold = None
    for e in T.all_edges:
        v1 = e.vertex1
        v2 = e.vertex2
        assert(e.vertex(0) == v1)
        assert(e.vertex(1) == v2)
        elen = e.length
        print(e, v1.index, v2.index, elen, e.midpoint, e.center)
        assert(e == e)
        if eold is not None:
            assert(e != eold)
        if e.is_infinite():
            assert(np.isclose(elen, -1.0))
            assert(np.isinf(e.center).all())
        else:
            p1 = e.point(0)
            p2 = e.point(1)
            l = np.sqrt(np.sum((p1 - p2)**2.0))
            assert(np.isclose(elen, l))
        eold = e


def test_cell():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    cold = None
    for c in T.all_cells:
        print(c, c.dimension, c.circumcenter, c.center)
        assert(c == c)
        if cold is not None:
            assert(c != cold)
        v1 = c.vertex(0)
        v2 = c.vertex(1)
        v3 = c.vertex(2)
        e1 = c.edge(0)
        e2 = c.edge(1)
        e3 = c.edge(2)
        assert(c.has_vertex(v1))
        assert(c.has_vertex(v1, return_index=True) == 0)
        assert(c.ind_vertex(v1) == 0)
        del e1, e2, e3

        c.reset_vertices()
        c.set_vertex(0, v1)
        c.set_vertices(v3, v2, v1)

        n1 = c.neighbor(0)
        n2 = c.neighbor(1)
        n3 = c.neighbor(2)
        assert(c.has_neighbor(n1))
        assert(c.has_neighbor(n1, return_index=True) == 0)
        assert(c.ind_neighbor(n1) == 0)

        c.reset_neighbors()
        c.set_neighbor(0, n1)
        c.set_neighbors(n3, n2, n1)

        p = c.circumcenter + 1.5*(c.point(0) - c.circumcenter)

        def rad(point, center):
            return np.sqrt(np.sum((point - center)**2))

        def p_outside(point, center):
            r_out = rad(point, center)
            p_out = center+1.1*np.sqrt(2)*r_out
            return p_out
        for i in range(3):
            print(i)
            print('    Cell: ', c.point(i), c.periodic_point(i),
                  c.periodic_offset(i))
            print('    Vert: ', c.vertex(i).point, c.vertex(i).periodic_point,
                  c.vertex(i).periodic_offset)
        print(p, rad(p, c.circumcenter), c.side_of_circle(p))
        print(80*'-')
        # print(c.side(c.center))
        # print(c.side(c.point(0)), c.side(v1.point), c.side(v2.point))
        # print(c.side(2*c.point(0)-c.center),
        #       c.side(2*v1.point-c.center),
        #       c.side(2*v2.point-c.center),
        #       c.side(p_outside(c.point(0), c.center)))
        # print(c.side_of_circle(c.circumcenter))
        # print(c.side_of_circle(c.point(0)), c.side_of_circle(v1.point))
        print(c.side_of_circle(1.01*c.point(0)-c.circumcenter),
              1.1*c.point(0)-c.circumcenter)
        print(c.side_of_circle(1.01*v1.point-c.circumcenter),
              1.1*v1.point-c.circumcenter)
        print(c.side_of_circle(p_outside(c.point(0), c.circumcenter)),
              p_outside(c.point(0), c.circumcenter))
        if c.is_infinite():
            assert(np.isinf(c.center).all())
            assert(c.side(c.center) == -1)
            assert(c.side(v1.point) == -1)
            assert(c.side(2*v1.point-c.center) == -1)
            assert(np.isinf(c.circumcenter).all())
            assert(c.side_of_circle(c.circumcenter) == -1)
            assert(c.side_of_circle(v1.point) == -1)
            assert(c.side_of_circle(2*v1.point-c.circumcenter) == -1)
        else:
            assert(c.side(c.center) == 1)
            if np.all((v1.point % 1) == 0):
                assert(c.side(v1.point) == 0)
            assert(c.side(p_outside(c.point(0), c.center)) == -1)
            assert(c.side_of_circle(c.circumcenter) == 1)
            if np.all((c.point(0) % 1) == 0):
                assert(c.side_of_circle(c.point(0)) == 0)
            # TODO: Fix this, currently fails
            # assert(c.side_of_circle(
            #            p_outside(c.point(0), c.circumcenter)) == -1)

        c.reorient()
        c.ccw_permute()
        c.cw_permute()

        cold = c


def test_move():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    v0 = T.get_vertex(0)
    new_pos = np.zeros(2, 'float64')
    v = T.move(v0, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(np.allclose(v0.point, new_pos))
    v1 = T.get_vertex(1)
    v = T.move(v1, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(T.num_verts == (nverts - 1))


def test_move_if_no_collision():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    v0 = T.get_vertex(0)
    new_pos = np.zeros(2, 'float64')
    v = T.move_if_no_collision(v0, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(np.allclose(v0.point, new_pos))
    v1 = T.get_vertex(1)
    v = T.move_if_no_collision(v1, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(np.allclose(v1.point, pts[1, :]))
    assert(T.num_verts == nverts)


def test_flip():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    for c in T.all_cells:
        out = T.flip(c, 0)
        assert(out)
    print(T.num_edges, nedges)
    assert(T.num_edges == nedges)
    for e in T.all_edges:
        out = e.flip()
        assert(out)
    print(T.num_edges, nedges)


def test_flippable():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    for c in T.all_cells:
        T.flip(c, 0)
    print(T.num_edges, nedges)
    assert(T.num_edges == nedges)
    for e in T.all_edges:
        e.flip()
    print(T.num_edges, nedges)


def test_vert_incident_verts():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for c in v.incident_vertices():
            c0 += 1
            count += 1
        print(v.index, c0)
    print(count)
    assert(count == 432)


def test_vert_incident_edges():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for e in v.incident_edges():
            c0 += 1
            count += 1
        print(v.index, c0)
    print(count)
    assert(count == 432)


def test_vert_incident_cells():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for c in v.incident_cells():
            c0 += 1
            count += 1
        print(v.index, c0)
    print(count)
    assert(count == 432)


def test_edge_incident_verts():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for v in T.all_edges:
        c0 = 0
        for e in v.incident_vertices():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 432)


def test_edge_incident_edges():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for v in T.all_edges:
        c0 = 0
        for e in v.incident_edges():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 2340)


def test_edge_incident_cells():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for v in T.all_edges:
        c0 = 0
        for e in v.incident_cells():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 2340)


def test_cell_incident_verts():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_vertices():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 432)


def test_cell_incident_edges():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_edges():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 432)


def test_cell_incident_cells():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_cells():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 432)


def test_nearest_vertex():
    idx_test = 7
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    v = T.nearest_vertex(pts[idx_test, :]-0.1)
    assert(v.index == idx_test)


def test_mirror():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    for c in T.all_cells:
        idx = 0
        i2 = T.mirror_index(c, idx)
        assert(c == c.neighbor(idx).neighbor(i2))
        v2 = T.mirror_vertex(c, idx)
        assert(v2 == c.neighbor(idx).vertex(i2))


def test_get_boundary_of_conflicts():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    v = T.get_vertex(0)
    c = v.incident_cells()[0]
    p = c.circumcenter
    edges = T.get_boundary_of_conflicts(p, c)
    print(len(edges))


def test_get_conflicts():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    v = T.get_vertex(0)
    c = v.incident_cells()[0]
    p = c.circumcenter
    cells = T.get_conflicts(p, c)
    print(len(cells))


def test_get_conflicts_and_boundary():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    v = T.get_vertex(0)
    c = v.incident_cells()[0]
    p = c.circumcenter
    cells, edges = T.get_conflicts_and_boundary(p, c)
    print(len(cells), len(edges))


def test_vertices():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    v = T.vertices
    assert(v.shape[0] == pts.shape[0])
    assert(v.shape[1] == pts.shape[1])
    assert(np.allclose(pts, v))


def test_edges():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    e = T.edges
    assert(e.shape[0] == T.num_finite_edges)
    assert(e.shape[1] == 2)


def test_boundary_points():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    out = T.boundary_points(left_edge, right_edge, False)
    del out


def test_outgoing_points():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    le = np.array([[-2, -2],
                   [-1, -1],
                   [+0,  0],
                   [+1,  1]], 'float64')
    re = np.array([[-1, -1],
                   [+0,  0],
                   [+1,  1],
                   [+2,  2]], 'float64')
    out = T.outgoing_points(le, re)
    assert(len(out) == le.shape[0])


def test_voronoi_volumes():
    T = Delaunay2(left_edge, right_edge)
    T.insert(pts)
    v = T.voronoi_volumes()
    assert(v.shape[0] == T.num_finite_verts)
    assert(np.all(v > 0))
    # TODO: answer testing



cimport numpy as np
from libcpp.vector cimport vector
from libcpp.set cimport set as cset
from libcpp.pair cimport pair
from libcpp cimport bool
from libc.stdint cimport uint32_t, uint64_t, int32_t, int64_t

cdef extern from "c_delaunayD.hpp":
    cdef int VALID
    cdef int D

    cdef cppclass Delaunay_with_info_D[Info] nogil:
        Delaunay_with_info_D() except +
        Delaunay_with_info_D(double *pts, Info *val, uint32_t n) except +
        bool updated
        bool is_valid() const
        uint32_t num_dims() const 
        uint32_t num_finite_verts() const
        uint32_t num_finite_cells() const
        uint32_t num_finite_faces(int d)
        uint32_t num_infinite_verts() const
        uint32_t num_infinite_cells() const
        uint32_t num_infinite_faces(int d)
        uint32_t num_verts() const
        uint32_t num_cells() const
        uint32_t num_faces(int d)

        bool is_equal(const Delaunay_with_info_D[Info] other) const

        cppclass Vertex
        cppclass Face
        cppclass Facet
        cppclass Cell

        Vertex infinite_vertex() const

        void insert(double *, Info *val, uint32_t n) except +
        void remove(Vertex) except +
        void clear() except + 

        # void write_to_file(const char* filename) except +
        # void read_from_file(const char* filename) except +
        I serialize[I](I &n, I &m, int32_t &d,
                       double* vert_pos, Info* vert_info,
                       I* cells, I* neighbors) const
        Info serialize_idxinfo[I](I &n, I &m, int32_t &d,
                                  Info* cells, I* neighbors) const
        I serialize_info2idx[I](I &n, I &m, int32_t &d,
                                I* faces, I* neighbors,
                                Info max_info, I* idx) const
        void deserialize[I](I n, I m, int32_t d,
                         double* vert_pos, Info* vert_info,
                         I* cells, I* neighbors, I idx_inf)
        void deserialize_idxinfo[I](I n, I m, int32_t d, double* vert_pos,
                                    I* cells, I* neighbors, I idx_inf)

        Vertex get_vertex(Info index) except +
        Cell locate(double* pos, int& lt, Face &f, Facet &ft)
        Cell locate(double* pos, int& lt, Face &f, Facet &ft, Cell c)

        void info_ordered_vertices(double* pos)
        void vertex_info(Info* verts)

        cppclass All_verts_iter:
            All_verts_iter()
            All_verts_iter& operator++()
            All_verts_iter& operator--()
            bool operator==(All_verts_iter other)
            bool operator!=(All_verts_iter other)
            Vertex vertex()
        All_verts_iter all_verts_begin()
        All_verts_iter all_verts_end()

        cppclass Vertex:
            Vertex()
            Vertex(All_verts_iter v)
            bool operator==(Vertex other)
            bool operator!=(Vertex other)
            void point(double* out)
            vector[double] point()
            Info info()
            Cell cell()
            void set_point(double* x)
            void set_cell(Cell c)

        cppclass Face:
            Face()
            # bool operator==(Face other)
            # bool operator!=(Face other)
            Vertex vertex(int i)
            int ind(int i)
            Cell cell()
            int dim()

        cppclass Facet:
            Facet()
            Facet(Cell x, int i1)
            # bool operator==(Facet other)
            # bool operator!=(Facet other)
            Cell cell()
            int ind()
            Vertex vertex(int i)

        cppclass All_cells_iter:
            All_cells_iter()
            All_cells_iter& operator++()
            All_cells_iter& operator--()
            bool operator==(All_cells_iter other)
            bool operator!=(All_cells_iter other)
        All_cells_iter all_cells_begin()
        All_cells_iter all_cells_end()

        cppclass Cell:
            Cell()
            Cell(All_cells_iter c)
            bool operator==(Cell other)
            bool operator!=(Cell other)

            Vertex vertex(int i)
            bool has_vertex(Vertex v)
            bool has_vertex(Vertex v, int *i)
            int ind(Vertex v)

            Cell neighbor(int i)
            bool has_neighbor(Cell c)
            bool has_neighbor(Cell c, int *i)
            int ind(Cell c)

            void set_vertex(int i, Vertex v)
            void set_neighbor(int i, Cell c)

        bool are_equal(Face f1, Face f2)
        bool are_equal(Facet f1, Facet f2)

        bool is_infinite(Vertex x)
        bool is_infinite(Face x)
        bool is_infinite(Facet x)
        bool is_infinite(Cell x)
        bool is_infinite(All_verts_iter x)
        bool is_infinite(All_cells_iter x)

        vector[Vertex] incident_vertices(Vertex x)
        vector[Vertex] incident_vertices(Face x)
        vector[Vertex] incident_vertices(Cell x)

        vector[Face] incident_faces(Vertex x, int i)
        vector[Face] incident_faces(Face x, int i)
        vector[Face] incident_faces(Cell x, int i)

        vector[Cell] incident_cells(Vertex x)
        vector[Cell] incident_cells(Face x)
        vector[Cell] incident_cells(Cell x)

        int mirror_index(Cell x, int i) const
        Vertex mirror_vertex(Cell x, int i) const

        void circumcenter(Cell x, double* out) const
        double n_simplex_volume(Face f) const
        double n_simplex_volume(Facet f) const
        double dual_volume(const Vertex v)
        void dual_volumes(double* vols)

        vector[vector[Info]] outgoing_points(uint64_t nbox,
                                             double *left_edges, double *right_edges)
